# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/7/1 11:24
describe: 期货CTA投研，支持按板块统计表现
"""
import os
import czsc
import json
import glob
import hashlib
import pandas as pd
import streamlit as st
import plotly.express as px
from pathlib import Path
from tqdm import tqdm
from typing import List
from loguru import logger
from czsc import subtract_fee, net_value_stats
from multiprocessing import cpu_count
from czsc.connectors.research import get_symbols, get_raw_bars
from czsc import CzscStrategyBase, Position


class JsonStreamStrategy(CzscStrategyBase):
    """读取 streamlit 传入的 json 策略，进行回测"""
    @property
    def positions(self) -> List[Position]:
        """返回当前的持仓策略"""
        json_strategies = self.kwargs.get("json_strategies")
        assert json_strategies, "请在初始化策略时，传入参数 json_strategies"
        positions = []
        for _, pos in json_strategies.items():
            pos["symbol"] = self.symbol
            positions.append(Position.load(pos))
        return positions


st.set_page_config(layout="wide", page_title="期货CTA投研", page_icon="🏕️")

future_plates = {
    "股指": ['SFIH9001', 'SFIC9001', 'SFIF9001', 'SFIM9001'],
    "黑色金属": ['ZZSF9001', 'ZZSM9001', 'SQhc9001', 'DLi9001', 'SQrb9001', 'SQss9001'],
    "轻工": ['SQsp9001', 'ZZFG9001', 'SEnr9001'],
    "软商品": ['ZZCF9001', 'ZZSR9001'],
    "贵金属": ['SQag9001', 'SQau9001'],
    "谷物": ['DLa9001', 'DLc9001'],
    "煤炭": ['DLj9001', 'DLjm9001'],
    "油脂油料": ['DLp9001', 'ZZRM9001', 'DLm9001', 'ZZOI9001', 'DLy9001'],
    "有色金属": ['SQsn9001', 'SQni9001', 'SQcu9001', 'SQzn9001', 'SQpb9001', 'SQal9001', 'SEbc9001'],
    "原油": ['SQfu9001', 'SEsc9001', 'DLpg9001', 'SElu9001'],
    "化工": [
        'SQbu9001',
        'DLeg9001',
        'ZZMA9001',
        'SQru9001',
        'DLl9001',
        'DLv9001',
        'ZZTA9001',
        'DLpp9001',
        'ZZUR9001',
        'DLeb9001',
        'ZZSA9001',
        'ZZPF9001',
    ],
    "农副": ['DLjd9001', 'DLlh9001', 'ZZAP9001', 'ZZPK9001'],
}

page_params = {"data_path": Path(r"D:\czsc_bi_datas\期货CTA投研")}
page_params['data_path'].mkdir(exist_ok=True, parents=True)


with st.sidebar:
    form = st.form(key='my_form_cta')
    files = form.file_uploader(label='上传策略文件', type='json', accept_multiple_files=True, key="files_cta")
    bar_sdt = form.date_input(label='行情开始日期', value=pd.to_datetime('2018-01-01'), key="bar_sdt_cta")
    sdt = form.date_input(label='回测开始日期', value=pd.to_datetime('2019-01-01'), key="sdt_cta")
    edt = form.date_input(label='回测结束日期', value=pd.to_datetime('2022-01-01'), key="edt_cta")
    max_workers = form.number_input(
        label='指定进程数量', value=cpu_count() // 4, min_value=1, max_value=cpu_count() // 2, key="max_workers_cta"
    )
    fee = int(form.number_input(label='单边手续费（单位：BP）', value=2, min_value=0, max_value=100, key="fee_cta"))
    submit_button = form.form_submit_button(label='开始回测')


@st.cache_data()
def read_holds_and_pairs(files_traders, pos_name, fee=1):
    holds, pairs = [], []
    for file in tqdm(files_traders):
        try:
            trader = czsc.dill_load(file)
            pos = trader.get_position(pos_name)
            if not pos.holds:
                logger.info(f"{trader.symbol} {pos_name} 无持仓，跳过")
                continue

            hd = pd.DataFrame(pos.holds)
            hd['symbol'] = trader.symbol
            hd = subtract_fee(hd, fee=fee)
            holds.append(hd)

            pr = pd.DataFrame(pos.pairs)
            pairs.append(pr)
        except Exception as e:
            logger.warning(f"{file} {pos_name} 读取失败: {e}")

    dfh = pd.concat(holds, ignore_index=True)
    dfp = pd.concat(pairs, ignore_index=True)
    return dfh, dfp


@st.cache_data()
def get_daily_nv(df):
    """获取每日净值"""
    res = []
    for symbol, hd in tqdm(df.groupby('symbol')):
        hd = hd.sort_values('dt', ascending=True)
        try:
            daily = hd.groupby('date').agg({'edge_pre_fee': 'sum', 'edge_post_fee': 'sum'}).reset_index()
            daily['symbol'] = symbol
            res.append(daily)
        except Exception as e:
            logger.exception(f"{symbol} 日收益获取失败: {e}")

    dfr = pd.concat(res, ignore_index=True)
    return dfr


def show_pos_detail(file_trader, pos_name):
    """显示持仓策略详情"""
    trader = czsc.dill_load(file_trader)
    pos = trader.get_position(pos_name)
    with st.expander(f"{pos_name} 持仓策略详情", expanded=False):
        _pos = pos.dump()
        _pos.pop('symbol')
        st.json(_pos)


def show_traders(file_traders, pos_name, fee=1):
    dfh, dfp = read_holds_and_pairs(file_traders, pos_name, fee=fee)
    dfr = get_daily_nv(dfh)
    show_pos_detail(file_traders[0], pos_name)

    st.subheader("一、单笔收益评价")
    from czsc import PairsPerformance

    pp = PairsPerformance(dfp)
    # st.write(pp.basic_info)
    df1 = pp.agg_statistics('标的代码')
    _res = pp.basic_info
    _res['标的代码'] = "全部品种"
    df1 = pd.concat([pd.DataFrame([_res]), df1], ignore_index=True)
    _cols = [
        '标的代码',
        '开始时间',
        '结束时间',
        '交易标的数量',
        '总体交易次数',
        '平均持仓K线数',
        '平均单笔收益',
        '单笔收益标准差',
        '交易胜率',
        '单笔盈亏比',
        '累计盈亏比',
        '盈亏平衡点',
        '每根K线收益',
    ]
    df1 = df1[_cols].set_index('标的代码')
    color_cols = ['交易标的数量', '总体交易次数', '平均持仓K线数', '平均单笔收益', '单笔收益标准差', 
                  '交易胜率', '单笔盈亏比', '累计盈亏比', '盈亏平衡点', '每根K线收益']
    df1 = df1.style.format('{0:,.2f}', subset=color_cols, na_rep="-").background_gradient(cmap='RdYlGn_r', subset=color_cols)

    st.dataframe(df1, use_container_width=True)

    st.divider()

    st.subheader("二、品种等权收益曲线")
    dfg = dfr.groupby('date').agg({'edge_pre_fee': 'mean', 'edge_post_fee': 'mean'}).cumsum()
    dfg.rename({'edge_pre_fee': '等权费前收益', 'edge_post_fee': f'双边扣费{2*fee}BP'}, axis=1, inplace=True)

    fig = px.line(dfg, x=dfg.index, y=['等权费前收益', f'双边扣费{2*fee}BP'], labels=[], title="全部品种日收益等权")
    st.plotly_chart(fig, use_container_width=True, height=600)

    dfg['dt'] = dfg.index.to_list()
    stats = []
    for col in ['等权费前收益', f'双边扣费{2*fee}BP']:
        dfg_ = dfg[['dt', col]].copy().rename(columns={col: 'edge'}).reset_index(drop=True)
        dfg_['edge'] = dfg_['edge'].diff()
        stats_ = net_value_stats(dfg_, sub_cost=False)
        stats_['name'] = col
        stats.append(stats_)
    st.dataframe(pd.DataFrame(stats).set_index('name'), use_container_width=True)


if files and sdt and max_workers:
    strategies = {file.name: json.loads(file.getvalue().decode("utf-8")) for file in files}
    symbols = get_symbols("期货主力")
    hash_code = hashlib.sha256(f"{str(strategies)}_{str(symbols)}".encode('utf-8')).hexdigest()[:8].upper()
    results_path = page_params['data_path'] / f"{sdt}_{edt}_{hash_code}"

    with st.sidebar.expander("策略详情", expanded=False):
        tactic = JsonStreamStrategy(json_strategies=strategies, symbol='symbol')
        st.caption(f"K线周期列表：{tactic.freqs}")
        st.caption("独立信号列表：")
        st.json(tactic.unique_signals)
        st.caption("信号函数配置：")
        st.json(tactic.signals_config)

    if not os.path.exists(results_path):
        os.makedirs(results_path, exist_ok=True)
        params = {"sdt": str(sdt), "edt": str(edt), "symbols": symbols}
        czsc.save_json(params, os.path.join(results_path, "params.json"))

        cta = czsc.CTAResearch(
            JsonStreamStrategy,
            get_raw_bars,
            results_path=results_path,
            json_strategies=strategies,
            signals_module_name='czsc.signals',
        )
        with st.spinner('正在执行策略回测，请耐心等候 ...'):
            cta.backtest(symbols, max_workers=int(max_workers), bar_sdt=bar_sdt, sdt=sdt, edt=edt)

    tabs = st.tabs(["所有品种", "行业板块"])
    with tabs[0]:
        file_traders = glob.glob(fr"{results_path}\backtest_*\traders\*.trader")
        all_pos_names = [x.name for x in czsc.dill_load(file_traders[0]).positions]
        pos_name = st.selectbox("选择持仓", all_pos_names, index=0, key="pos_name")
        show_traders(file_traders, pos_name, fee=fee)

    with tabs[1]:
        col1, col2 = st.columns([2, 4])
        plate = col1.selectbox("选择板块", list(future_plates.keys()), index=0, key="plate")
        symbols = future_plates[plate]  # type: ignore
        st.caption(f"板块包含品种：{symbols}")
        file_traders = glob.glob(fr"{results_path}\backtest_*\traders\*.trader")
        file_traders = [x for x in file_traders if os.path.basename(x).split(".")[0] in symbols]
        all_pos_names = [x.name for x in czsc.dill_load(file_traders[0]).positions]
        pos_name = col2.selectbox("选择持仓", all_pos_names, index=0, key="plate_pos_name")
        show_traders(file_traders, pos_name, fee=fee)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/10/24 16:20
describe: 测试 Tushare 数据缓存机制
"""
import sys
sys.path.insert(0, '.')
sys.path.insert(0, '..')

import os
from czsc.data.ts_cache import *

os.environ['czsc_verbose'] = '1'
data_path = r"C:\ts_data_czsc"


def test_ts_cache_ths_daily():
    dc = TsDataCache(data_path=data_path, sdt='20200101')
    df = dc.ths_daily('885573.TI', start_date="20200101", end_date="20220517", raw_bar=False)
    assert df.shape[0] == 572 and df.shape[1] == 34
    df = dc.ths_daily('885573.TI', start_date="20200101", end_date="20220420", raw_bar=False)
    assert df.shape[0] == 556 and df.shape[1] == 34

    # 测试被动刷新数据
    dc = TsDataCache(data_path=data_path, sdt='20210101')
    df = dc.ths_daily('885573.TI', start_date="20210101", end_date="20220517", raw_bar=False)
    assert df.shape[0] == 329 and df.shape[1] == 34

    # 测试主动刷新数据
    dc = TsDataCache(data_path=data_path, refresh=True, sdt='20210101')
    df = dc.ths_daily('885573.TI', start_date="20210101", end_date="20220517", raw_bar=False)
    assert df.shape[0] == 329 and df.shape[1] == 34


def test_ts_cache_daily_basic_new():
    dc = TsDataCache(data_path=data_path, sdt='20200101', edt='20211024')
    df = dc.daily_basic_new(trade_date='2018-03-15')
    assert df.shape[0] == 3237 and df.shape[1] == 37

    dfb = dc.stocks_daily_basic_new(sdt='20211001', edt='20211020')
    assert dfb.shape[1] == df.shape[1] + 1 and len(dfb) == 40407


def test_ts_cache_bars():
    """测试获取K线"""
    dc = TsDataCache(data_path=data_path, sdt='20200101', edt='20211024')
    # 测试日线以上数据获取
    bars = dc.pro_bar(ts_code='000001.SZ', asset='E', freq='D',
                      start_date='20200101', end_date='20211024', raw_bar=True)
    assert len(bars) == 436
    df = dc.pro_bar(ts_code='000001.SZ', asset='E', freq='D',
                    start_date='20200101', end_date='20211024', raw_bar=False)
    assert len(df) == 436
    df = dc.pro_bar(ts_code='000001.SZ', asset='E', freq='D',
                    start_date='20210108', end_date='20210108', raw_bar=False)
    assert len(df) == 1

    # 测试日线主动刷新
    dc = TsDataCache(data_path=data_path, refresh=True, sdt='20200101', edt='20211024')
    bars = dc.pro_bar(ts_code='000001.SZ', asset='E', freq='D',
                      start_date='20200101', end_date='20211024', raw_bar=True)
    assert len(bars) == 436

    # 测试日线被动刷新
    dc = TsDataCache(data_path=data_path, refresh=False, sdt='20210101', edt='20211024')
    bars = dc.pro_bar(ts_code='000001.SZ', asset='E', freq='D',
                      start_date='20210101', end_date='20211024', raw_bar=True)
    assert len(bars) == 193


def test_pro_bar_minutes():
    # 测试复权分钟线获取
    dc = TsDataCache(data_path=data_path, sdt='20200101', edt='20211024')
    df1 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj='hfq', raw_bar=False)
    bars1 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                               sdt="20200101", edt="20210804 11:24", adj='qfq', raw_bar=True)
    df2 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj='qfq', raw_bar=False)
    df3 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj=None, raw_bar=False)
    assert len(df1) == len(df2) == len(df3) == len(bars1) \
           and df1.iloc[-1]['close'] > df3.iloc[-1]['close'] > df2.iloc[-1]['close']

    # 测试获取指数分钟行情
    df1 = dc.pro_bar_minutes(ts_code='000001.SZ', asset='I', freq='30min',
                             sdt="20200101", edt="20210804 11:24", raw_bar=False)
    assert len(df1) == 3083

    # 测试 ETF 复权行情
    df1 = dc.pro_bar_minutes(ts_code='512880.SH', asset='FD', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj='qfq', raw_bar=False)
    assert round(df1.iloc[-1]['close'], 4) == 1.066
    assert len(df1) == 3083

    # 测试主动刷新
    dc = TsDataCache(data_path=data_path, refresh=True, sdt='20200101', edt='20211024')
    df1 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj='hfq', raw_bar=False)
    assert len(df1) == 3083

    # 测试被动刷新
    dc = TsDataCache(data_path=data_path, refresh=False, sdt='20190101', edt='20211024')
    df1 = dc.pro_bar_minutes(ts_code='000002.SZ', asset='E', freq='30min',
                             sdt="20200101", edt="20210804 11:24", adj='hfq', raw_bar=False)
    assert len(df1) == 3083


def test_ts_cache():
    dc = TsDataCache(data_path=data_path, sdt='20200101', edt='20211024')
    assert dc.get_next_trade_dates('2022-03-02', 2, 5) == ['20220304', '20220307', '20220308']
    assert dc.get_next_trade_dates('2022-03-02', -1, -4) == ['20220224', '20220225', '20220228']
    assert dc.get_dates_span('20220224', '20220228') == ['20220224', '20220225', '20220228']
    assert dc.get_dates_span('20220224', '20220228', is_open=False) \
           == ['20220224', '20220225', '20220226', '20220227', '20220228']

    hk_holds = dc.hk_hold('20211103')
    assert hk_holds.shape[0] == 2965
    hk_holds = dc.hk_hold('20211103')
    assert hk_holds.shape[0] == 2965

    news = dc.cctv_news('20211103')
    assert news.shape[0] == 14
    news = dc.cctv_news('20211103')
    assert news.shape[0] == 14

    # 测试指数成分和权重数据缓存
    df = dc.index_weight('000905.SH', '20210923')
    assert len(df) == 502
    df = dc.index_weight('000905.SH', '20210901')
    assert len(df) == 502
    df = dc.index_weight('000300.SH', '20200208')
    assert len(df) == 300

    df = dc.get_all_ths_members(exchange='A', type_='N')
    assert not df.empty

    df = dc.limit_list(trade_date='20210324')
    assert not df.empty and os.path.exists(os.path.join(dc.api_path_map['limit_list'], "limit_list_20210324.feather"))

    x1 = dc.get_next_trade_dates('2021-12-13', n=-1, m=None)
    assert x1 == '20211210'
    x1 = dc.get_next_trade_dates('2021-12-13', n=1, m=None)
    assert x1 == '20211214'
    x2 = dc.get_next_trade_dates('2021-12-13', n=1, m=6)
    assert x2 == ['20211214', '20211215', '20211216', '20211217', '20211220']

    df = dc.ths_index('A', 'N')
    assert not df.empty and os.path.exists(os.path.join(dc.api_path_map['ths_index'], "ths_index_A_N.feather"))
    df = dc.ths_index('A')
    assert not df.empty
    df = dc.ths_index('A', None)
    assert not df.empty and os.path.exists(os.path.join(dc.api_path_map['ths_index'], "ths_index_A_None.feather"))

    df = dc.get_all_ths_members('A', None)
    assert not df.empty and os.path.exists(os.path.join(dc.cache_path, "A_None_ths_members.feather"))

    bars = dc.ths_daily(ts_code='885566.TI', start_date='20200101', end_date='20211024', raw_bar=False)
    assert len(bars) == 436



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 21:43
describe: 概念、行业、指数等股票聚类板块感应器

三个核心问题：
1）如何找出引领大盘的概念、行业、指数
2）板块内股票相比于板块走势，划分强中弱
3）根据指数强弱进行账户总仓位控制
"""
import os
import pandas as pd
from tqdm import tqdm
from typing import Callable
from czsc.utils.plt_plot import plot_bins_return
from czsc.data.ts_cache import TsDataCache
from czsc.sensors.utils import discretizer


class MeanPlatesSensor:
    """基于股票打分取平均的板块轮动观察

    数据要素：
    1. 全市场所有股票日线行情
    2. 板块列表，板块成分股，所有板块日线行情
    """
    def __init__(self,
                 dc: TsDataCache,
                 res_path,
                 sdt,
                 mean_col,
                 mean_col_bins,
                 sort_col,
                 sort_col_bins,
                 create_features: Callable,
                 get_plates: Callable,
                 max_num=100,
                 max_adj=10):
        """

        :param dc:
        :param res_path:
        :param sdt: 开始时间
        :param mean_col: 个股上用于计算板块均值的特征列
        :param mean_col_bins: 对mean_col按分位数分成20层，指定选中哪些分位数区间
        :param sort_col: 个股上用于排序约束组合换手率的特征列
        :param sort_col_bins: 对sort_col按分位数分成20层，指定选中哪些分位数区间
        :param create_features: 个股特征创建函数
        :param get_plates: 板块划分的数据获取函数
        """
        self.dc = dc
        self.sdt = sdt
        self.mean_col = mean_col
        self.mean_col_bins = mean_col_bins
        self.sort_col = sort_col
        self.sort_col_bins = sort_col_bins
        self.res_path = res_path
        self.create_features = create_features
        self.get_plates = get_plates
        os.makedirs(self.res_path, exist_ok=True)

        # 创建实验结果目录
        self.exp_path = os.path.join(res_path, f"{get_plates.__name__}_{mean_col}_{sdt}")
        os.makedirs(self.exp_path, exist_ok=True)

        # 获取全市场股票相关信息
        file_dfb = os.path.join(self.res_path, 'dfb.pkl')
        if os.path.exists(file_dfb):
            self.dfb = pd.read_pickle(file_dfb)
        else:
            # 统一用后复权数据处理，实盘跟踪过程也使用后复权
            dfb = dc.stocks_daily_bars(sdt=dc.sdt, edt=dc.edt, adj='hfq')
            dfb = self.create_features(dfb)
            # 去除ST，引入市值之类的指标
            dfc = dc.stocks_daily_basic_new(dc.sdt, dc.edt)
            dfb = dfb.merge(dfc, on=['ts_code', 'trade_date'], how='left', suffixes=("", "_drop"))
            dfb.drop([x for x in dfb.columns if x.endswith("_drop")], axis=1, inplace=True)
            self.dfb = dfb[(dfb['dt'] >= pd.to_datetime(sdt)) & (dfb['is_st'] == False)]
            self.dfb.to_pickle(file_dfb)

        self.n_cols = [x for x in self.dfb.columns if x[0] == 'n' and x[-1] == 'b']
        self.b_cols = [x for x in self.dfb.columns if x[0] == 'b' and x[-1] == 'b']

        # p 是 plates 的首字母，代表板块列表；p_bars 所有板块的日线行情；p_members 所有板块的成分股
        self.p, self.p_bars, self.p_members = self.get_plates()

        file_plates = os.path.join(self.res_path, f'plates_{mean_col}.pkl')
        if os.path.exists(file_plates):
            self.plates = pd.read_pickle(file_plates)
        else:
            self.plates = self.mean_by_plates()
            self.plates.to_pickle(file_plates)

        # 执行组合构建
        self.dfh1 = self.create_base_holds(mean_col, mean_col_bins)
        self.dfh2 = self.create_holds(mean_col, mean_col_bins, sort_col, sort_col_bins, max_num, max_adj)

    def mean_by_plates(self):
        """计算每个板块每天的打分"""
        sdt = self.dfb['dt'].min().strftime("%Y%m%d")
        mean_col = self.mean_col
        dc = self.dc
        # 全市场所有股票行情特征按 dt 聚合
        df_map = {dt_: dfg for dt_, dfg in self.dfb.dropna(subset=[mean_col]).groupby('dt')}

        # 按板块聚合成分股
        pm_map = {k: dfg for k, dfg in self.p_members.groupby('ts_code')}
        p = [x for x in self.p.to_dict('records') if x['ts_code'] in pm_map.keys()]

        _results = []
        for date in tqdm(dc.get_dates_span(sdt, dc.edt), desc=f"mean_by_plates | {mean_col}"):
            try:
                dft = df_map[pd.to_datetime(date)]
                for _row in p:
                    members = pm_map[_row['ts_code']]
                    dfm = dft[dft['ts_code'].isin(members['code'])]
                    _row[mean_col] = dfm[mean_col].mean() if len(dfm) > 0 else 0

                _df = pd.DataFrame(p).sort_values(mean_col, ascending=False).reset_index(drop=True)
                _df['dt'] = date
                _results.append(_df)
            except:
                print(f"fail on {date}")

        _df_plates = pd.concat(_results, ignore_index=True)
        _df_plates['trade_date'] = pd.to_datetime(_df_plates['dt'])
        df_plates = _df_plates.merge(self.p_bars, on=['ts_code', 'trade_date'], how='left')
        df_plates['dt'] = pd.to_datetime(df_plates['trade_date'])

        df_plates = discretizer(df_plates, mean_col, n_bins=20)
        file_png = os.path.join(self.res_path, f"{mean_col}_bins20_plates.png")
        plot_bins_return(df_plates, f"{mean_col}_bins20", file_png)
        return df_plates

    def get_base_stocks(self, mean_col, mean_col_bins, sort_col=None, sort_col_bins=None):
        """结合板块轮动和个股RPS打分构建基础持仓组合"""
        df_plates = self.plates
        dfb = self.dfb
        p_members = self.p_members
        df_sp = df_plates[df_plates[f"{mean_col}_bins20"].isin(mean_col_bins)]

        _results = []
        for dt, dfg in df_sp.groupby('dt'):
            # 获取 dfg 中给定板块的所有成分股
            df_tm = p_members[p_members['ts_code'].isin(dfg['ts_code'])]
            # 暂时不选北交所股票
            selected = [x for x in df_tm['code'].unique().tolist() if x.split('.')[1] in ['SH', 'SZ']]
            _dfs = pd.DataFrame({'dt': dt, 'ts_code': selected})
            _results.append(_dfs)
        dfs = pd.concat(_results)
        dfs = dfs.merge(dfb, on=['dt', 'ts_code'], how='left')
        dfs['下期一字板'] = dfs['下期一字板'].fillna(0)
        dfs = dfs[(dfs['下期一字板'] == 0) & (dfs['is_st'] != True)]

        # dt 后延一期，因为 T 时刻选出的股票是 T+1 时刻的持仓
        dts = list(pd.to_datetime(self.dc.get_dates_span('20000101', '20300101')))
        next_dt = {dts[i]: dts[i+1] for i in range(len(dts)-1)}
        dfs['dt'] = dfs['dt'].apply(lambda x: next_dt[x])

        if sort_col and sort_col_bins:
            dfs = dfs[dfs[f"{sort_col}_bins20"].isin(sort_col_bins)]
        return dfs

    def create_base_holds(self, mean_col, mean_col_bins):
        """选取板块轮动的所有成分股构建持仓基础持仓组合"""
        dfs = self.get_base_stocks(mean_col, mean_col_bins, sort_col=None, sort_col_bins=None)
        holds = []
        dfs_map = {date: dfg for date, dfg in dfs.groupby('trade_date')}
        dates = sorted(dfs_map.keys())

        for date in tqdm(dates, desc=f"create_base_holds"):
            _df = dfs_map[date]
            _dfh = _df[['dt', 'ts_code', 'n1b']].reset_index(drop=True)
            _dfh['持仓权重'] = 1 / len(_dfh)
            _dfh.rename({'dt': '成分日期', 'ts_code': '证券代码'}, axis=1, inplace=True)
            holds.append(_dfh)

        dfh = pd.concat(holds, ignore_index=True)
        dfh['成分日期'] = dfh['成分日期'].apply(lambda x: x.strftime("%Y-%m-%d"))

        exp_name = f"P{'#'.join([str(x) for x in mean_col_bins])}"
        res_path = os.path.join(self.exp_path, exp_name)

        from czsc.traders.performance import stock_holds_performance
        stock_holds_performance(self.dc, dfh, res_path)
        return dfh

    def create_holds(self, mean_col, mean_col_bins, sort_col, sort_col_bins, max_num=100, max_adj=10):
        """结合板块轮动和个股RPS打分构建持仓组合

        :param mean_col: 板块轮动列
        :param mean_col_bins: 板块轮动列分层可取值
        :param sort_col: 个股RPS打分列
        :param sort_col_bins: 个股RPS打分列分层可取值
        :param max_num: 组合最大持仓数量约束
        :param max_adj: 组合每期可调整数量约束
        :return:
        """
        dfs = self.get_base_stocks(mean_col, mean_col_bins, sort_col, sort_col_bins)
        dfb = self.dfb

        # 约束最大持仓数量和每天最多换仓比例
        holds = []
        last = pd.DataFrame()
        dfs_map = {date: dfg for date, dfg in dfs.groupby('trade_date')}
        dates = sorted(dfs_map.keys())
        weight = 1 / max_num

        for date in tqdm(dates, desc=f"create_holds | {sort_col}"):
            _df = dfs_map[date]

            if last.empty:
                # 交易第一天直接买满 max_num
                assert not holds
                _df = _df.sort_values([sort_col], ascending=False).head(max_num)
            else:
                # 删掉已经有的持仓股
                _df = _df[~_df['ts_code'].isin(last['证券代码'])]

                # 获取可以加入组合的 max_adj 只股票
                _df = _df.sort_values([sort_col], ascending=False).head(max_adj)

            _dfh = _df[['dt', 'ts_code', 'n1b']].reset_index(drop=True)
            _dfh['持仓权重'] = weight
            _dfh.rename({'dt': '成分日期', 'ts_code': '证券代码'}, axis=1, inplace=True)

            if last.empty:
                hold = _dfh.copy(deep=True)
            else:
                # 根据 RPS 删掉 max_adj 只原有持仓
                last['成分日期'] = date
                last = last[['成分日期', '证券代码', '持仓权重']]
                last = last.merge(dfb[['dt', 'ts_code', sort_col, 'n1b']],
                                  left_on=['成分日期', '证券代码'], right_on=['dt', 'ts_code'], how='left')
                last = last.sort_values([sort_col], ascending=False).head(max_num - max_adj)
                last = last[['成分日期', '证券代码', '持仓权重', 'n1b']]

                # 更新持仓
                hold = pd.concat([last, _dfh], ignore_index=True)

            holds.append(hold)
            last = hold.copy(deep=True)
        dfh = pd.concat(holds, ignore_index=True)
        dfh['成分日期'] = dfh['成分日期'].apply(lambda x: x.strftime("%Y-%m-%d"))

        exp_name = f"P{'#'.join([str(x) for x in mean_col_bins])}_" \
                   f"{sort_col}{'#'.join([str(x) for x in sort_col_bins])}_{max_num}_{max_adj}"
        res_path = os.path.join(self.exp_path, exp_name)
        from czsc.traders.performance import stock_holds_performance
        stock_holds_performance(self.dc, dfh, res_path=res_path)
        return dfh



# 创建K线分割规则
import pandas as pd
from czsc.connectors.research import get_raw_bars


def init_kline():
    bars = get_raw_bars(symbol="000001.SH", freq="1分钟", sdt="20210903", edt="20210904", fq="后复权")
    dft = pd.DataFrame(bars)
    dft['time'] = dft['dt'].dt.strftime('%H:%M')
    dft[['time']].to_excel('test.xlsx')


    bars = get_raw_bars(symbol="SQau9001", freq="1分钟", sdt="20210903", edt="20210904", fq="后复权")
    dft = pd.DataFrame(bars)
    dft['time'] = dft['dt'].dt.strftime('%H:%M')
    dft[['time']].to_excel('test.xlsx')


# df = pd.read_excel(r"C:\Users\zengb\Desktop\time_split_conf_V2.xlsx")
# df['time_t'] = pd.to_datetime(df['time'], format='%H:%M')

# # for i in (3, 5, 10, 15, 20, 30, 60):
# #     df[f"{i}分钟A"] = df['time_t'].apply(lambda x: x + pd.Timedelta(minutes=i - x.minute % i))
# #     df[f"{i}分钟A"] = df[f"{i}分钟A"].dt.strftime('%H:%M').shift(1)

# # df['20分钟A'] = df['20分钟'].fillna(method='bfill')
# # for i in (2, 4, 6, 12):
# #     df[f"{i}分钟A"] = df['time_t'].apply(lambda x: x + pd.Timedelta(minutes=i - x.minute % i))
# #     df[f"{i}分钟A"] = df[f"{i}分钟A"].dt.strftime('%H:%M').shift(1)

# # df.to_excel(r"C:\Users\zengb\Desktop\time_split_conf_V2.xlsx", index=False)


# df['60分钟'] = df['60分钟'].fillna(method='bfill')
# df['120分钟'] = df['120分钟'].fillna(method='bfill')
# df.to_excel(r"C:\Users\zengb\Desktop\time_split_conf_V3.xlsx", index=False)

# df = pd.read_excel(r"C:\Users\zengb\Desktop\time_split_conf_V3.xlsx")
# df.to_feather("minites_split.feather")

# 默认分割规则

def split_time(freq="60分钟"):
    sdt = pd.to_datetime("2021-09-03 00:00")
    res = [sdt]
    for i in range(2000):
        sdt += pd.Timedelta(minutes=1)
        res.append(sdt)
    df = pd.DataFrame(res, columns=['time'])
    df['t1'] = df['time'].dt.strftime('%H:%M')
    df1 = df.resample(freq.replace("分钟", "T"), on='time').last().reset_index()
    df1['t2'] = df1['time'].dt.strftime('%H:%M')
    dfx = pd.merge_asof(df, df1, on='time', direction='forward')
    dfx = dfx.dropna().copy()
    dfx['time'] = dfx['t1_x']
    dfx['edt'] = dfx['t2']
    dfx['freq'] = freq
    dfx = dfx[['time', 'edt', 'freq']].drop_duplicates().sort_values('time')
    return dfx

rows = []
for freq in ("1分钟", "2分钟", "3分钟", "4分钟", "5分钟", "6分钟", "10分钟", "12分钟", "15分钟", "20分钟", "30分钟", "60分钟", '120分钟'):
    rows.append(split_time(freq))

df = pd.concat(rows, ignore_index=True)
df = pd.pivot_table(df, index='time', columns='freq', values='edt', aggfunc='first').reset_index()
df['market'] = "默认"
df = df[['market', 'time', '1分钟', '2分钟', '3分钟', '4分钟', '5分钟', '6分钟', '10分钟', '12分钟', '15分钟', '20分钟', '30分钟', '60分钟', '120分钟']]
df.to_excel(r"C:\Users\zengb\Desktop\time_split_conf_V4.xlsx", index=False)

df = pd.read_excel(r"C:\Users\zengb\Desktop\time_split_conf_V4.xlsx")
df.to_feather(r"D:\ZB\git_repo\waditu\czsc\czsc\utils\minites_split.feather")


# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/9/24 15:19
describe: 策略持仓权重管理
"""
import os
import time
import json
import redis
import threading
import pandas as pd
from loguru import logger
from datetime import datetime


class RedisWeightsClient:
    """策略持仓权重收发客户端"""

    version = "V231112"

    def __init__(self, strategy_name, redis_url=None, send_heartbeat=True, **kwargs):
        """
        :param strategy_name: str, 策略名
        :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取

            For example::

                redis://[[username]:[password]]@localhost:6379/0
                rediss://[[username]:[password]]@localhost:6379/0
                unix://[username@]/path/to/socket.sock?db=0[&password=password]

            Three URL schemes are supported:

            - `redis://` creates a TCP socket connection. See more at:
            <https://www.iana.org/assignments/uri-schemes/prov/redis>
            - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
            <https://www.iana.org/assignments/uri-schemes/prov/rediss>
            - ``unix://``: creates a Unix Domain Socket connection.

        :param send_heartbeat: boolean, 是否发送心跳

            如果为True，会在后台启动一个线程，每15秒向redis发送一次心跳，用于检测策略是否存活。
            推荐在写入数据时设置为True，读取数据时设置为False，避免无用的心跳。

        :param kwargs: dict, 其他参数

            - key_prefix: str, redis中key的前缀，默认为 Weights
            - heartbeat_prefix: str, 心跳key的前缀，默认为 heartbeat
        """
        self.strategy_name = strategy_name
        self.redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
        self.key_prefix = kwargs.get("key_prefix", "Weights")

        thread_safe_pool = redis.BlockingConnectionPool.from_url(self.redis_url, decode_responses=True)
        self.r = redis.Redis(connection_pool=thread_safe_pool)
        self.lua_publish = RedisWeightsClient.register_lua_publish(self.r)

        if send_heartbeat:
            self.heartbeat_client = redis.from_url(self.redis_url, decode_responses=True)
            self.heartbeat_prefix = kwargs.get("heartbeat_prefix", "heartbeat")
            self.heartbeat_thread = threading.Thread(target=self.__heartbeat, daemon=True)
            self.heartbeat_thread.start()

    def set_metadata(self, base_freq, description, author, outsample_sdt, **kwargs):
        """设置策略元数据"""
        key = f'{self.key_prefix}:META:{self.strategy_name}'
        if self.r.exists(key):
            if not kwargs.pop('overwrite', False):
                logger.warning(f'已存在 {self.strategy_name} 的元数据，如需覆盖请设置 overwrite=True')
                return
            else:
                self.r.delete(key)
                logger.warning(f'删除 {self.strategy_name} 的元数据，重新写入')

        outsample_sdt = pd.to_datetime(outsample_sdt).strftime('%Y%m%d')
        meta = {'name': self.strategy_name, 'base_freq': base_freq, 'key_prefix': self.key_prefix,
                'description': description, 'author': author, 'outsample_sdt': outsample_sdt,
                'update_time': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'kwargs': json.dumps(kwargs)}
        self.r.hset(key, mapping=meta)

    @property
    def metadata(self):
        """获取策略元数据"""
        key = f'{self.key_prefix}:META:{self.strategy_name}'
        return self.r.hgetall(key)

    def get_last_times(self, symbols=None):
        """获取所有品种上策略最近一次发布信号的时间

        :param symbols: list, 品种列表, 默认为None, 即获取所有品种
        :return: dict, {symbol: datetime}，如{'SFIF9001': datetime(2021, 9, 24, 15, 19, 0)}
        """
        if isinstance(symbols, str):
            row = self.r.hgetall(f'{self.key_prefix}:{self.strategy_name}:{symbols}:LAST')
            return pd.to_datetime(row['dt']) if row else None  # type: ignore

        symbols = symbols if symbols else self.get_symbols()
        with self.r.pipeline() as pipe:
            for symbol in symbols:
                pipe.hgetall(f'{self.key_prefix}:{self.strategy_name}:{symbol}:LAST')
            rows = pipe.execute()
        return {x['symbol']: pd.to_datetime(x['dt']) for x in rows}

    def publish(self, symbol, dt, weight, price=0, ref=None, overwrite=False):
        """发布单个策略持仓权重

        :param symbol: str, eg; SFIF9001
        :param dt: py_datetime or pandas Timestamp
        :param weight: float, 信号值
        :param price: float, 产生信号时的价格
        :param ref: dict, 自定义数据
        :param overwrite: boolean, 是否覆盖已有记录
        :return: 成功发布信号的条数
        """
        if not isinstance(dt, datetime):
            dt = pd.to_datetime(dt)

        if not overwrite:
            last_dt = self.get_last_times(symbol)
            if last_dt is not None and dt <= last_dt:   # type: ignore
                logger.warning(f"不允许重复写入，已过滤 {symbol} {dt} 的重复信号")
                return 0

        udt = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        key = f'{self.key_prefix}:{self.strategy_name}:{symbol}:{dt.strftime("%Y%m%d%H%M%S")}'
        ref = ref if ref else '{}'
        ref_str = json.dumps(ref) if isinstance(ref, dict) else ref
        return self.lua_publish(keys=[key], args=[1 if overwrite else 0, udt, weight, price, ref_str])

    def publish_dataframe(self, df, overwrite=False, batch_size=10000):
        """批量发布多个策略信号

        :param df: pandas.DataFrame, 必需包含['symbol', 'dt', 'weight']列,
                可选['price', 'ref']列, 如没有price则写0, dtype同publish方法
        :param overwrite: boolean, 是否覆盖已有记录
        :return: 成功发布信号的条数
        """
        df = df.copy()
        df['dt'] = pd.to_datetime(df['dt'])
        logger.info(f"输入数据中有 {len(df)} 条权重信号")

        # 去除单个品种下相邻时间权重相同的数据
        _res = []
        for _, dfg in df.groupby('symbol'):
            dfg = dfg.sort_values('dt', ascending=True).reset_index(drop=True)
            dfg = dfg[dfg['weight'].diff().fillna(1) != 0].copy()
            _res.append(dfg)
        df = pd.concat(_res, ignore_index=True)
        df = df.sort_values(['dt']).reset_index(drop=True)
        logger.info(f"去除单个品种下相邻时间权重相同的数据后，剩余 {len(df)} 条权重信号")

        if 'price' not in df.columns:
            df['price'] = 0
        if 'ref' not in df.columns:
            df['ref'] = '{}'

        if not overwrite:
            raw_count = len(df)
            _time = self.get_last_times()
            _data = []
            for symbol, dfg in df.groupby('symbol'):
                last_dt = _time.get(symbol)
                if last_dt is not None:
                    dfg = dfg[dfg['dt'] > last_dt]
                _data.append(dfg)
            df = pd.concat(_data, ignore_index=True)
            logger.info(f"不允许重复写入，已过滤 {raw_count - len(df)} 条重复信号")

        keys, args = [], []
        for row in df[['symbol', 'dt', 'weight', 'price', 'ref']].to_numpy():
            key = f'{self.key_prefix}:{self.strategy_name}:{row[0]}:{row[1].strftime("%Y%m%d%H%M%S")}'
            keys.append(key)

            args.append(row[2])
            args.append(row[3])
            ref = row[4]
            args.append(json.dumps(ref) if isinstance(ref, dict) else ref)

        udt = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        overwrite = 1 if overwrite else 0

        pub_cnt = 0
        len_keys = len(keys)
        for i in range(0, len_keys, batch_size):
            if i + batch_size < len_keys:
                tmp_keys = keys[i: i + batch_size]
                tmp_args = [overwrite, udt] + args[3 * i: 3 * (i + batch_size)]
            else:
                tmp_keys = keys[i: len_keys]
                tmp_args = [overwrite, udt] + args[3 * i: 3 * len_keys]
            logger.info(f"索引 {i}，即将发布 {len(tmp_keys)} 条权重信号")
            pub_cnt += self.lua_publish(keys=tmp_keys, args=tmp_args)
            logger.info(f"已完成 {pub_cnt} 次发布")
        return pub_cnt

    def __heartbeat(self):
        while True:
            key = f'{self.key_prefix}:{self.heartbeat_prefix}:{self.strategy_name}'
            try:
                self.heartbeat_client.set(key, datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
            except Exception:
                continue
            time.sleep(15)

    def get_keys(self, pattern):
        """获取 redis 中指定 pattern 的 keys"""
        return self.r.keys(pattern)

    def clear_all(self):
        """删除该策略所有记录"""
        self.r.delete(f'{self.key_prefix}:META:{self.strategy_name}')
        keys = self.get_keys(f'{self.key_prefix}:{self.strategy_name}*')
        if keys is not None and len(keys) > 0:  # type: ignore
            self.r.delete(*keys)                # type: ignore

    @staticmethod
    def register_lua_publish(client):
        lua_body = '''
local overwrite = ARGV[1]
local update_time = ARGV[2]
local cnt = 0
local ret
for i = 1, #KEYS do
    local key = KEYS[i]
    local sig = ARGV[3 + 3 * (i - 1)]
    local price = ARGV[4 + 3 * (i - 1)]
    local ref_str = ARGV[5 + 3 * (i - 1)]
    local split_str = {}
    key:gsub('[^:]+', function(s) table.insert(split_str, s) end)
    local model_key = split_str[1] .. ':' .. split_str[2] .. ':' .. split_str[3]

    local if_pass = true
    if overwrite ~= '0' then
        if_pass = false
    else
        local pos = redis.call('HGET', model_key .. ':LAST', 'weight')
        if not pos or math.abs(tonumber(sig) - tonumber(pos)) > 0.00001 then
            if_pass = false
        end
    end

    if not if_pass then
        local strategy_name, symbol, action_time = split_str[2], split_str[3], split_str[4]
        local at_str = string.sub(action_time, 1, 4) .. '-' .. string.sub(action_time, 5, 6) .. '-' ..
            string.sub(action_time, 7, 8) .. ' ' .. string.sub(action_time, 9, 10) .. ':' ..
            string.sub(action_time, 11, 12) .. ':' .. string.sub(action_time, 13, 14)
        redis.call('ZADD', model_key, tonumber(action_time), key)
        local ret1 = redis.call('HMSET', key, 'symbol', symbol, 'weight', sig, 'dt', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        local ret2 = redis.call('HMSET', key:gsub(action_time, 'LAST'), 'symbol', symbol, 'weight', sig, 'dt', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        if ret1.ok and ret2.ok then
            cnt = cnt + 1
            local pubKey = 'PUBSUB:' .. split_str[1] .. ':' .. strategy_name .. ':' .. symbol
            redis.call('PUBLISH', pubKey, key .. ':' .. sig .. ':' .. price .. ':' .. ref_str)
        end
    end
end
return cnt
'''
        return client.register_script(lua_body)

    def get_symbols(self):
        """获取策略交易的品种列表"""
        keys = self.get_keys(f'{self.key_prefix}:{self.strategy_name}*')
        symbols = {x.split(":")[2] for x in keys}       # type: ignore
        return list(symbols)

    def get_last_weights(self, symbols=None, ignore_zero=True, lua=True):
        """获取最近的持仓权重

        :param symbols: list, 品种列表
        :param ignore_zero: boolean, 是否忽略权重为0的品种
        :param lua: boolean, 是否使用 lua 脚本获取，默认为True
            如果要全量获取，推荐使用 lua 脚本，速度更快；如果要获取指定 symbols，不推荐使用 lua 脚本。
        :return: pd.DataFrame
        """
        if lua:
            lua_script = """
            local keys = redis.call('KEYS', ARGV[1])
            local results = {}
            for i=1, #keys do
                results[i] = redis.call('HGETALL', keys[i])
            end
            return results
            """
            key_pattern = self.key_prefix + ':' + self.strategy_name + ':*:LAST'
            results = self.r.eval(lua_script, 0, key_pattern)
            rows = [dict(zip(r[::2], r[1::2])) for r in results]     # type: ignore
            if symbols:
                rows = [r for r in rows if r['symbol'] in symbols]

        else:
            symbols = symbols if symbols else self.get_symbols()
            with self.r.pipeline() as pipe:
                for symbol in symbols:
                    pipe.hgetall(f'{self.key_prefix}:{self.strategy_name}:{symbol}:LAST')
                rows = pipe.execute()

        dfw = pd.DataFrame(rows)
        dfw['weight'] = dfw['weight'].astype(float)
        dfw['dt'] = pd.to_datetime(dfw['dt'])
        if ignore_zero:
            dfw = dfw[dfw['weight'] != 0].copy().reset_index(drop=True)
        dfw = dfw.sort_values(['dt', 'symbol']).reset_index(drop=True)
        return dfw

    def get_hist_weights(self, symbol, sdt, edt) -> pd.DataFrame:
        """获取单个品种的持仓权重历史数据

        :param symbol: str, 品种代码
        :param sdt: str, 开始时间, eg: 20210924 10:19:00
        :param edt: str, 结束时间, eg: 20220924 10:19:00
        :return: pd.DataFrame
        """
        start_score = pd.to_datetime(sdt).strftime('%Y%m%d%H%M%S')
        end_score = pd.to_datetime(edt).strftime('%Y%m%d%H%M%S')
        model_key = f'{self.key_prefix}:{self.strategy_name}:{symbol}'
        key_list = self.r.zrangebyscore(model_key, start_score, end_score)

        if len(key_list) == 0:
            logger.warning(f'no history weights: {symbol} - {sdt} - {edt}')
            return pd.DataFrame()

        with self.r.pipeline() as pipe:
            for key in key_list:
                pipe.hmget(key, 'weight', 'price', 'ref')
            rows = pipe.execute()

        weights = []
        for i in range(len(key_list)):
            dt = pd.to_datetime(key_list[i].split(":")[-1])
            weight, price, ref = rows[i]
            weight = weight if weight is None else float(weight)
            price = price if price is None else float(price)
            try:
                ref = json.loads(ref)
            except Exception:
                ref = ref
            weights.append((self.strategy_name, symbol, dt, weight, price, ref))

        dfw = pd.DataFrame(weights, columns=['strategy_name', 'symbol', 'dt', 'weight', 'price', 'ref'])
        dfw = dfw.sort_values('dt').reset_index(drop=True)
        return dfw

    def get_all_weights(self, sdt=None, edt=None, ignore_zero=True) -> pd.DataFrame:
        """获取所有权重数据

        :param sdt: str, 开始时间, eg: 20210924 10:19:00
        :param edt: str, 结束时间, eg: 20220924 10:19:00
        :param ignore_zero: boolean, 是否忽略权重为0的品种
        :return: pd.DataFrame
        """
        lua_script = """
        local keys = redis.call('KEYS', ARGV[1])
        local results = {}
        for i=1, #keys do
            local last_part = keys[i]:match('([^:]+)$')
            if #last_part == 14 and tonumber(last_part) ~= nil then
                results[#results + 1] = redis.call('HGETALL', keys[i])
            end
        end
        return results
        """
        key_pattern = self.key_prefix + ':' + self.strategy_name + ':*:*'
        results = self.r.eval(lua_script, 0, key_pattern)
        results = [dict(zip(r[::2], r[1::2])) for r in results]     # type: ignore

        df = pd.DataFrame(results)
        df['dt'] = pd.to_datetime(df['dt'])
        df['weight'] = df['weight'].astype(float)
        df = df.sort_values(['dt', 'symbol']).reset_index(drop=True)

        df1 = pd.pivot_table(df, index='dt', columns='symbol', values='weight').sort_index().ffill().fillna(0)
        df1 = pd.melt(df1.reset_index(), id_vars='dt', value_vars=df1.columns, value_name='weight')     # type: ignore
        if ignore_zero:
            df1 = df1[df1['weight'] != 0].reset_index(drop=True)
        if sdt:
            df1 = df1[df1['dt'] >= pd.to_datetime(sdt)].reset_index(drop=True)
        if edt:
            df1 = df1[df1['dt'] <= pd.to_datetime(edt)].reset_index(drop=True)
        df1 = df1.sort_values(['dt', 'symbol']).reset_index(drop=True)
        return df1



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/12/24 22:20
describe: 简单的单仓位策略执行
"""
import os
import webbrowser
import numpy as np
import pandas as pd
from tqdm import tqdm
from loguru import logger
from datetime import datetime, timedelta
from deprecated import deprecated
from collections import OrderedDict
from typing import Callable, List, AnyStr, Union, Optional
from pyecharts.charts import Tab
from pyecharts.components import Table
from pyecharts.options import ComponentTitleOpts
from czsc.analyze import CZSC
from czsc.objects import Position, RawBar, Signal
from czsc.utils.bar_generator import BarGenerator
from czsc.utils.cache import home_path
from czsc.utils import sorted_freqs, import_by_name
from czsc.traders.sig_parse import get_signals_freqs


class CzscSignals:
    """缠中说禅技术分析理论之多级别信号计算"""

    def __init__(self, bg: Optional[BarGenerator] = None, **kwargs):
        """

        :param bg: K线合成器
        """
        self.name = "CzscSignals"
        # cache 是信号计算过程的缓存容器，需要信号计算函数自行维护
        self.cache = OrderedDict()
        self.kwargs = kwargs
        self.signals_config = kwargs.get("signals_config", [])

        if bg:
            self.bg = bg
            assert bg.symbol, "bg.symbol is None"
            self.symbol = bg.symbol
            self.base_freq = bg.base_freq
            self.freqs = list(bg.bars.keys())
            self.kas = {freq: CZSC(b) for freq, b in bg.bars.items()}

            last_bar = self.kas[self.base_freq].bars_raw[-1]
            self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
            self.s = OrderedDict()
            self.s.update(self.get_signals_by_conf())
            self.s.update(last_bar.__dict__)
        else:
            self.bg = None
            self.symbol = None
            self.base_freq = None
            self.freqs = None
            self.kas = None
            self.end_dt, self.bid, self.latest_price = None, None, None
            self.s = OrderedDict()

    def __repr__(self):
        return "<{} for {}>".format(self.name, self.symbol)

    def get_signals_by_conf(self):
        """通过信号参数配置获取信号

        信号参数配置，格式如下：

            signals_config = [
                {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 1, 'ma_type': 'SMA', 'timeperiod': 5},
                {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 5, 'ma_type': 'SMA', 'timeperiod': 5},
                {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 1, 'ma_seq': (5, 20), 'th': 100},
                {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 5, 'ma_seq': (5, 20), 'th': 100},
            ]

        :return: 信号字典
        """
        s = OrderedDict()
        if not self.signals_config:
            return s

        for param in self.signals_config:
            param = dict(param)
            sig_name = param.pop('name')
            sig_func = import_by_name(sig_name) if isinstance(sig_name, str) else sig_name

            freq = param.pop('freq', None)
            if freq in self.kas:    # 如果指定了 freq，那么就使用 CZSC 对象作为输入
                s.update(sig_func(self.kas[freq], **param))
            else:                   # 否则使用 CAT 作为输入
                s.update(sig_func(self, **param))
        return s

    def take_snapshot(self, file_html=None, width: str = "1400px", height: str = "580px"):
        """获取快照

        :param file_html: 交易快照保存的 html 文件名
        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            ka: CZSC = self.kas[freq]
            chart = ka.to_echarts(width, height)
            tab.add(chart, freq)

        signals = {k: v for k, v in self.s.items() if len(k.split("_")) == 3}
        for freq in self.freqs:
            # 按各周期K线分别加入信号表
            freq_signals = {k: signals[k] for k in signals.keys() if k.startswith("{}_".format(freq))}
            for k in freq_signals.keys():
                signals.pop(k)
            if len(freq_signals) <= 0:
                continue
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in freq_signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, f"{freq}信号")

        if len(signals) > 0:
            # 加入时间、持仓状态之类的其他信号
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, "其他信号")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def open_in_browser(self, width="1400px", height="580px"):
        """直接在浏览器中打开分析结果"""
        file_html = os.path.join(home_path, "temp_czsc_advanced_trader.html")
        self.take_snapshot(file_html, width, height)
        webbrowser.open(file_html)

    def update_signals(self, bar: RawBar):
        """输入基础周期已完成K线，更新信号，更新仓位"""
        self.bg.update(bar)
        for freq, b in self.bg.bars.items():
            self.kas[freq].update(b[-1])

        self.symbol = bar.symbol
        last_bar = self.kas[self.base_freq].bars_raw[-1]
        self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
        self.s = OrderedDict()
        self.s.update(self.get_signals_by_conf())
        self.s.update(last_bar.__dict__)


def generate_czsc_signals(bars: List[RawBar], signals_config: List[dict],
                          sdt: Union[AnyStr, datetime] = "20170101", init_n: int = 500, df=False, **kwargs):
    """使用 CzscSignals 生成信号

    :param bars: 基础周期 K 线序列
    :param signals_config: 信号函数配置，格式如下：
        signals_config = [
            {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 1, 'ma_type': 'SMA', 'timeperiod': 5},
            {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 5, 'ma_type': 'SMA', 'timeperiod': 5},
            {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 1, 'ma_seq': (5, 20), 'th': 100},
            {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 5, 'ma_seq': (5, 20), 'th': 100},
        ]
    :param sdt: 信号计算开始时间
    :param init_n: 用于 BarGenerator 初始化的基础周期K线数量
    :param df: 是否返回 df 格式的信号计算结果，默认 False
    :return: 信号计算结果
    """
    freqs = get_signals_freqs(signals_config)
    freqs = [freq for freq in freqs if freq != bars[0].freq.value]
    sdt = pd.to_datetime(sdt)                       # type: ignore
    bars_left = [x for x in bars if x.dt < sdt]     # type: ignore
    if len(bars_left) <= init_n:
        bars_left = bars[:init_n]
        bars_right = bars[init_n:]
    else:
        bars_right = [x for x in bars if x.dt >= sdt]   # type: ignore

    if len(bars_right) == 0:
        logger.warning("右侧K线为空，无法进行信号生成", category=RuntimeWarning)
        if df:
            return pd.DataFrame()
        else:
            return []

    base_freq = str(bars[0].freq.value)
    bg = BarGenerator(base_freq=base_freq, freqs=freqs, max_count=kwargs.get("bg_max_count", 5000))
    for bar in bars_left:
        bg.update(bar)

    _sigs = []
    cs = CzscSignals(bg, signals_config=signals_config, **kwargs)
    cs.cache.update({'gsc_kwargs': kwargs})
    for bar in tqdm(bars_right, desc=f'generate signals of {bg.symbol}'):
        cs.update_signals(bar)
        _sigs.append(dict(cs.s))

    if df:
        return pd.DataFrame(_sigs)
    else:
        return _sigs


def check_signals_acc(bars: List[RawBar], signals_config: List[dict], delta_days: int = 5, **kwargs) -> None:
    """人工验证形态信号识别的准确性的辅助工具：

    输入基础周期K线和想要验证的信号，输出信号识别结果的快照

    :param bars: 原始K线
    :param signals_config: 需要验证的信号列表
    :param delta_days: 两次相同信号之间的间隔天数
    :return: None
    """
    base_freq = str(bars[-1].freq.value)
    assert bars[2].dt > bars[1].dt > bars[0].dt and bars[2].id > bars[1].id, "bars 中的K线元素必须按时间升序"
    if len(bars) < 600:
        return

    df = generate_czsc_signals(bars, signals_config=signals_config, df=True, **kwargs)
    s_cols = [x for x in df.columns if len(x.split("_")) == 3]
    signals = []
    for col in s_cols:
        print('=' * 100, "\n", df[col].value_counts())
        signals.extend([Signal(f"{col}_{v}") for v in df[col].unique() if "其他" not in v])

    print(f"signals: {'+' * 100}")
    for row in signals:
        print(f"- {row}")

    bars_left = bars[:500]
    bars_right = bars[500:]
    freqs = get_signals_freqs(signals_config)
    bg = BarGenerator(base_freq=base_freq, freqs=freqs, max_count=5000)
    for bar in bars_left:
        bg.update(bar)

    ct = CzscSignals(bg, signals_config=signals_config, **kwargs)
    last_dt = {signal.key: ct.end_dt for signal in signals}

    for bar in tqdm(bars_right, desc=f'signals of {bg.symbol}'):
        ct.update_signals(bar)

        for signal in signals:
            html_path = os.path.join(home_path, signal.key)
            os.makedirs(html_path, exist_ok=True)
            if bar.dt - last_dt[signal.key] > timedelta(days=delta_days) and signal.is_match(ct.s):
                file_html = f"{bar.symbol}_{bar.dt.strftime('%Y%m%d_%H%M')}_{signal.key}_{ct.s[signal.key]}.html"
                file_html = os.path.join(html_path, file_html)
                print(file_html)
                ct.take_snapshot(file_html, height=kwargs.get("height", "680px"))
                last_dt[signal.key] = bar.dt


def get_unique_signals(bars: List[RawBar], signals_config: List[dict], **kwargs):
    """获取信号函数中定义的所有信号列表

    :param bars: 基础K线数据
    :param signals_config: 信号函数配置
    :param kwargs:
    :return:
    """
    assert bars[2].dt > bars[1].dt > bars[0].dt and bars[2].id > bars[1].id, "bars 中的K线元素必须按时间升序"
    if len(bars) < 600:
        return []

    df = generate_czsc_signals(bars, signals_config=signals_config, df=True, **kwargs)
    _res = []
    for col in [x for x in df.columns if len(x.split("_")) == 3]:
        _res.extend([f"{col}_{v}" for v in df[col].unique() if "其他" not in v])
    return _res


class CzscTrader(CzscSignals):
    """缠中说禅技术分析理论之多级别联立交易决策类（支持多策略独立执行）"""

    def __init__(self, bg: Optional[BarGenerator] = None, positions: Optional[List[Position]] = None,
                 ensemble_method: Union[AnyStr, Callable] = "mean", **kwargs):
        """

        :param bg: bar generator 对象
        :param get_signals: 信号计算函数，输入是 CzscSignals 对象，输出是信号字典
        :param ensemble_method: 多个仓位集成一个仓位的方法，可选值 mean, vote, max；也可以传入一个回调函数

            假设有三个仓位对象，当前仓位分别是 1, 1, -1
            mean - 平均仓位，pos = np.mean([1, 1, -1]) = 0.33
            vote - 投票表决，pos = 1
            max  - 取最大，pos = 1

            对于传入回调函数的情况，函数的输入为 dict，key 为 position.name，value 为 position.pos, 样例输入：
            {'多头策略A': 1, '多头策略B': 1, '空头策略A': -1}
        """
        self.positions = positions
        if self.positions:
            _pos_names = [x.name for x in self.positions]
            assert len(_pos_names) == len(set(_pos_names)), "仓位策略名称不能重复"
        self.__ensemble_method = ensemble_method
        self.name = "CzscTrader"
        super().__init__(bg, **kwargs)

    def __repr__(self):
        return "<{} for {}>".format(self.name, self.symbol)

    def update(self, bar: RawBar) -> None:
        """输入基础周期已完成K线，更新信号，更新仓位

        :param bar: 基础周期已完成K线
        :return: None
        """
        self.update_signals(bar)
        if self.positions:
            for position in self.positions:
                position.update(self.s)

    def on_sig(self, sig: dict) -> None:
        """通过信号字典直接交易

        主要用于快速回测场景

        :param sig: 信号字典
        :return: None
        """
        self.s = sig
        self.symbol, self.end_dt = self.s['symbol'], self.s['dt']
        self.bid, self.latest_price = self.s['id'], self.s['close']
        if self.positions:
            for position in self.positions:
                position.update(self.s)

    def on_bar(self, bar: RawBar) -> None:
        """输入基础周期已完成K线，更新信号，更新仓位

        :param bar: 基础周期已完成K线
        :return: None
        """
        self.update(bar)

    @property
    def pos_changed(self) -> bool:
        """判断仓位是否发生变化

        :return: True/False
        """
        if not self.positions:
            return False
        return any([position.pos_changed for position in self.positions])

    def get_ensemble_pos(self, method: Union[AnyStr, Callable] = None) -> float:
        """获取多个仓位的集成仓位

        :param method: 多个仓位集成一个仓位的方法，可选值 mean, vote, max；也可以传入一个回调函数

            假设有三个仓位对象，当前仓位分别是 1, 1, -1
            mean - 平均仓位，pos = np.mean([1, 1, -1]) = 0.33
            vote - 投票表决，pos = 1
            max  - 取最大，pos = 1

            对于传入回调函数的情况，输入是 self.positions

        :return: pos, 集成仓位
        """
        if not self.positions:
            return 0

        method = self.__ensemble_method if not method else method
        if isinstance(method, str):
            method = method.lower()
            pos_seq = [x.pos for x in self.positions]

            if method == 'mean':
                pos = np.mean(pos_seq)
            elif method == 'vote':
                pos = np.sign(sum(pos_seq))
            elif method == 'max':
                pos = max(pos_seq)
            else:
                raise ValueError

        else:
            pos = method({x.name: x.pos for x in self.positions})

        return pos

    def get_position(self, name: str) -> Optional[Position]:
        """获取指定名称的仓位策略对象

        :param name: 仓位名称
        :return: Position
        """
        if not self.positions:
            return None

        for position in self.positions:
            if position.name == name:
                return position

        return None

    def take_snapshot(self, file_html=None, width: str = "1400px", height: str = "580px"):
        """获取快照

        :param file_html: 交易快照保存的 html 文件名
        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            ka: CZSC = self.kas[freq]
            bs = None
            if freq == self.base_freq and self.positions:
                # 在基础周期K线上加入最近的操作记录
                bs = []
                for pos in self.positions:
                    for op in pos.operates:
                        if op['dt'] >= ka.bars_raw[0].dt:
                            _op = dict(op)
                            _op['op_desc'] = f"{pos.name} | {_op['op_desc']}"
                            bs.append(_op)

            chart = ka.to_echarts(width, height, bs)
            tab.add(chart, freq)

        signals = {k: v for k, v in self.s.items() if len(k.split("_")) == 3}
        for freq in self.freqs:
            # 按各周期K线分别加入信号表
            freq_signals = {k: signals[k] for k in signals.keys() if k.startswith("{}_".format(freq))}
            for k in freq_signals.keys():
                signals.pop(k)
            if len(freq_signals) <= 0:
                continue
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in freq_signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, f"{freq}信号")

        if len(signals) > 0:
            # 加入时间、持仓状态之类的其他信号
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, "其他信号")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def get_ensemble_weight(self, method: Optional[Union[AnyStr, Callable]] = None):
        """获取 CzscTrader 中所有 positions 按照 method 方法集成之后的权重

        :param method: str or callable
            集成方法，可选值包括：'mean', 'max', 'min', 'vote'
            也可以传入自定义的函数，函数的输入为 dict，key 为 position.name，value 为 position.pos, 样例输入：
                {'多头策略A': 1, '多头策略B': 1, '空头策略A': -1}
        :param kwargs:
        :return: pd.DataFrame
            columns = ['dt', 'symbol', 'weight', 'price']
        """
        from czsc.traders.weight_backtest import get_ensemble_weight
        method = self.__ensemble_method if not method else method
        return get_ensemble_weight(self, method)

    def weight_backtest(self, **kwargs):
        """执行仓位集成权重的回测

        :param kwargs:

            - method: str or callable，集成方法，参考 get_ensemble_weight 方法
            - digits: int，权重小数点后保留的位数，例如 2 表示保留两位小数
            - fee_rate: float，手续费率，例如 0.0002 表示万二
            - res_path: str，回测结果保存路径

        :return: 回测结果
        """
        from czsc.traders.weight_backtest import WeightBacktest

        method = kwargs.get("method", self.__ensemble_method)
        digits = kwargs.get("digits", 2)
        fee_rate = kwargs.get("fee_rate", 0.0002)
        res_path = kwargs.get("res_path", "./weight_backtest")
        dfw = self.get_ensemble_weight(method)
        wb = WeightBacktest(dfw, digits=digits, fee_rate=fee_rate, res_path=res_path)
        _res = wb.backtest()
        return _res



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/21 16:04
describe: 交易相关的工具函数
"""
import pandas as pd
from typing import List, Union
from czsc.objects import RawBar


def risk_free_returns(start_date="20180101", end_date="20210101", year_returns=0.03):
    """创建无风险收益率序列

    创建一个 Pandas DataFrame，包含两列："date" 和 "returns"。"date" 列包含从 trade_dates 获取的所有交易日期，
    "returns" 列包含无风险收益率序列，计算方法是将年化收益率（year_returns）除以 252（一年的交易日数量，假设为每周 5 天）

    :param start_date: 起始日期
    :param end_date: 截止日期
    :param year_returns: 年化收益率
    :return: pd.DataFrame
    """
    from czsc.utils.calendar import get_trading_dates
    trade_dates = get_trading_dates(start_date, end_date)   # type: ignore
    df = pd.DataFrame({"date": trade_dates, "returns": year_returns / 252})
    return df


def cal_trade_price(bars: Union[List[RawBar], pd.DataFrame], decimals=3, **kwargs):
    """计算给定品种基础周期K线数据的交易价格

    函数执行逻辑：

    1. 首先，根据输入的 bars 参数类型（列表或 DataFrame），将其转换为 DataFrame 格式，并将其存储在变量 df 中。
    2. 计算下一根K线的开盘价和收盘价，分别存储在新列 next_open 和 next_close 中。同时，将这两个新列名添加到 price_cols 列表中。
    3. 计算 TWAP（时间加权平均价格）和 VWAP（成交量加权平均价格）。为此，函数使用了一个 for 循环，
       遍历 t_seq 参数（默认值为 (5, 10, 15, 20, 30, 60)）。在每次循环中：

        - 计算 TWAP：使用 rolling(t).mean().shift(-t) 方法计算时间窗口为 t 的滚动平均收盘价。
        - 计算 VWAP：首先计算滚动窗口内的成交量之和（sum_vol_t）和成交量乘以收盘价之和（sum_vcp_t），然后用后者除以前者，并向下移动 t 个单位。
        - 将 TWAP 和 VWAP 的列名添加到 price_cols 列表中。

    4. 遍历 price_cols 列表中的每个列，将其中的 NaN 值替换为对应行的收盘价。
    5. 从 DataFrame 中选择所需的列（包括基本的K线数据列和新计算的交易价格列），并使用 round(decimals) 方法保留指定的小数位数（默认为3）。
    6. 返回处理后的 DataFrame。

    :param bars: 基础周期K线数据，一般是1分钟周期的K线
    :param decimals: 保留小数位数，默认值3
    :return: 交易价格表
    """
    df = pd.DataFrame(bars) if isinstance(bars, list) else bars

    # 下根K线开盘、收盘
    df['next_open'] = df['open'].shift(-1)
    df['next_close'] = df['close'].shift(-1)
    price_cols = ['next_open', 'next_close']

    # TWAP / VWAP 价格
    df['vol_close_prod'] = df['vol'] * df['close']
    for t in kwargs.get('t_seq', (5, 10, 15, 20, 30, 60)):
        df[f"TWAP{t}"] = df['close'].rolling(t).mean().shift(-t)
        df[f"sum_vol_{t}"] = df['vol'].rolling(t).sum()
        df[f"sum_vcp_{t}"] = df['vol_close_prod'].rolling(t).sum()
        df[f"VWAP{t}"] = (df[f"sum_vcp_{t}"] / df[f"sum_vol_{t}"]).shift(-t)
        price_cols.extend([f"TWAP{t}", f"VWAP{t}"])
        df.drop(columns=[f"sum_vol_{t}", f"sum_vcp_{t}"], inplace=True)

    df.drop(columns=['vol_close_prod'], inplace=True)
    # 用当前K线的收盘价填充交易价中的 nan 值
    for price_col in price_cols:
        df.loc[df[price_col].isnull(), price_col] = df[df[price_col].isnull()]['close']

    df[price_cols] = df[price_cols].round(decimals)
    return df


def update_nbars(da, price_col='close', numbers=(1, 2, 5, 10, 20, 30), move=0) -> None:
    """在给定的 da 上计算并添加后面 n 根 bar 的累计收益列

    收益计量单位：BP；1倍涨幅 = 10000BP

    函数的逻辑如下：

    1. 首先，检查 price_col 是否在输入的 DataFrame（da）的列名中。如果不在，抛出 ValueError。
    2. 确保 move 是一个非负整数。
    3. 使用 for 循环遍历 numbers 列表中的每个整数 n, 对于每个整数 n，计算 n 根 bar 的累计收益。
    4. 返回 None，表示这个函数会直接修改输入的 DataFrame（da），而不返回新的 DataFrame。

    :param da: 数据，DataFrame结构
    :param price_col: 价格列
    :param numbers: 考察的bar的数目的列表
    :param move: 收益计算是否要整体移位，move必须是非负整数
        一般是当前bar的close计算收益，也可以考虑是下根bar的open。这个时候 move=1。
    :return nbars_cols: 后面n根bar的bp值列名
    """
    if price_col not in da.columns:
        raise ValueError(f"price_col {price_col} not in da.columns")

    assert move >= 0
    for n in numbers:
        da[f"n{n}b"] = (da[price_col].shift(-n - move) / da[price_col].shift(-move) - 1) * 10000.0


def update_bbars(da, price_col='close', numbers=(1, 2, 5, 10, 20, 30)) -> None:
    """在给定的 da 数据上计算并添加前面 n 根 bar 的累计收益列

    函数的逻辑如下：

    1. 首先，检查 price_col 是否在输入的 DataFrame（da）的列名中。如果不在，抛出 ValueError。
    2. 使用 for 循环遍历 numbers 列表中的每个整数 n，对于每个整数 n，计算 n 根 bar 的累计收益。
    3. 返回 None，表示这个函数会直接修改输入的 da，而不返回新的 DataFrame。

    :param da: K线数据，DataFrame结构
    :param price_col: 价格列
    :param numbers: 考察的bar的数目的列表
    :return: bbars_cols: 后面n根bar的bp值列名
    """
    if price_col not in da.columns:
        raise ValueError(f"price_col {price_col} not in da.columns")

    for n in numbers:
        # 收益计量单位：BP；1倍涨幅 = 10000BP
        da[f"b{n}b"] = (da[price_col] / da[price_col].shift(n) - 1) * 10000


def update_tbars(da: pd.DataFrame, event_col: str) -> None:
    """计算带 Event 方向信息的未来收益

    函数的逻辑如下：

    1. 从输入的 da的列名中提取所有以 'n' 开头，以 'b' 结尾的列名，这些列名表示未来 n 根 bar 的累计收益。将这些列名存储在 n_seq 列表中。
    2. 使用 for 循环遍历 n_seq 列表中的每个整数 n。
    3. 对于每个整数 n，计算带有 Event 方向信息的未来收益。
        计算方法是：将前面 n 根 bar 的累计收益（列名 f'n{n}b'）与事件信号列（event_col）的值相乘。
        将计算结果存储在一个新的列中，列名为 f't{n}b'。
    4. 返回 None，表示这个函数会直接修改输入的 da，而不返回新的 DataFrame。

    :param da: K线数据，DataFrame结构
    :param event_col: 事件信号列名，含有 0, 1, -1 三种值，0 表示无事件，1 表示看多事件，-1 表示看空事件
    :return: None
    """
    n_seq = [int(x.strip('nb')) for x in da.columns if x[0] == 'n' and x[-1] == 'b']
    for n in n_seq:
        da[f't{n}b'] = da[f'n{n}b'] * da[event_col]



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/2/23 19:13
describe: 测试策略基类
"""
import os
import shutil
import pandas as pd
from test.test_analyze import read_1min
from czsc.utils.bar_generator import resample_bars, Freq
from czsc.traders import generate_czsc_signals


def test_czsc_strategy_example2():
    """测试策略示例2：仅传入Positions构建策略"""
    from czsc.strategies import CzscStrategyExample2

    bars = read_1min()
    df = pd.DataFrame(bars)
    bars = resample_bars(df, Freq.F15)

    strategy = CzscStrategyExample2(symbol="000001.SH")

    trader1 = strategy.init_trader(bars=bars, init_n=2000, sdt="20170101")
    assert len(trader1.positions) == 2
    sigs = generate_czsc_signals(bars, freqs=strategy.sorted_freqs,
                                 init_n=2000, sdt="20170101", signals_config=strategy.signals_config)
    trader2 = strategy.dummy(sigs)
    assert len(trader2.positions) == 2
    trader3 = strategy.dummy(sigs, sleep_time=0.1)
    assert len(trader3.positions) == 2

    for i in [0, 1]:
        pos1 = trader1.positions[i]
        assert trader1.positions[i].evaluate()['日胜率'] == trader2.positions[i].evaluate()['日胜率']
        assert len(trader1.positions[i].pairs) == len(trader2.positions[i].pairs)
        assert round(pos1.evaluate("多空")['覆盖率'], 3) == round(pos1.evaluate("多头")['覆盖率'] + pos1.evaluate("空头")['覆盖率'], 3)

    strategy.replay(bars, res_path="trade_replay_test", sdt='20190101', refresh=True)
    assert len(os.listdir("trade_replay_test")) == 3
    shutil.rmtree("trade_replay_test")

    # 验证信号计算的准确性
    strategy.check(bars, res_path="trade_check_test", sdt='20190101', exist_ok=False)
    assert len(os.listdir("trade_check_test")) == 2
    assert os.path.exists(os.path.join("trade_check_test", "signals.xlsx"))
    assert os.path.exists(os.path.join("trade_check_test", "15分钟_D0停顿分型_BE辅助V230106"))
    shutil.rmtree("trade_check_test")



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/17 18:50
"""
import pandas as pd
import numpy as np
from tqdm import tqdm
from collections import Counter
from typing import List
from sklearn.preprocessing import KBinsDiscretizer
from ..data import TsDataCache


def discretizer(df: pd.DataFrame, col: str, n_bins=20, encode='ordinal', strategy='quantile'):
    """使用 KBinsDiscretizer 对连续变量在时间截面上进行离散化

    :param df: 数据对象
    :param col: 连续变量列名
    :param n_bins: 参见 KBinsDiscretizer 文档
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KBinsDiscretizer.html
    :param encode: 参见 KBinsDiscretizer 文档
    :param strategy: 参见 KBinsDiscretizer 文档
    :return:
    """
    assert col in df.columns, f'{col} not in {df.columns}'
    assert 'dt' in df.columns

    new_col = f'{col}_bins{n_bins}'
    results = []
    for dt, dfg in tqdm(df.groupby('dt'), desc=f"{col}_bins{n_bins}"):
        kb = KBinsDiscretizer(n_bins=n_bins, encode=encode, strategy=strategy)
        # 加1，使分组从1开始
        dfg[new_col] = kb.fit_transform(dfg[col].values.reshape(-1, 1)).ravel() + 1
        results.append(dfg)
    df = pd.concat(results, ignore_index=True)
    return df


def get_index_beta(dc: TsDataCache, sdt: str, edt: str, freq='D', file_xlsx=None, indices=None):
    """获取基准指数的Beta

    :param dc: 数据缓存对象
    :param sdt: 开始日期
    :param edt: 结束日期
    :param freq: K线周期，D 日线，W 周线，M 月线
    :param file_xlsx: 结果保存文件
    :param indices: 定义指数列表
    :return:
    """
    if not indices:
        indices = ['000001.SH', '000016.SH', '000905.SH', '000300.SH', '399001.SZ', '399006.SZ']

    beta = {}
    p = []
    for ts_code in indices:
        df = dc.pro_bar(ts_code=ts_code, start_date=sdt, end_date=edt, freq=freq, asset="I", raw_bar=False)
        beta[ts_code] = df
        df = df.fillna(0)
        start_i, end_i, mdd = max_draw_down(df['n1b'].to_list())
        start_dt = df.iloc[start_i]['trade_date']
        end_dt = df.iloc[end_i]['trade_date']
        row = {
            '标的': ts_code,
            "开始日期": sdt,
            "结束日期": edt,
            "最大回撤": mdd,
            "回撤开始": start_dt,
            "回撤结束": end_dt,
            "交易次数": len(df),
            "交易胜率": round(len(df[df.n1b > 0]) / len(df), 4),
            "累计收益": round(df.n1b.sum(), 4),
        }
        cols = [x for x in df.columns if x[0] == 'n' and x[-1] == 'b']
        row.update({x: round(df[x].mean(), 4) for x in cols})
        p.append(row)

    dfp = pd.DataFrame(p)
    if file_xlsx:
        f = pd.ExcelWriter(file_xlsx)
        dfp.to_excel(f, index=False, sheet_name="指数表现")
        for name, df_ in beta.items():
            df_.to_excel(f, index=False, sheet_name=name)
        f.close()
    else:
        beta['dfp'] = dfp
        return beta


def max_draw_down(n1b: List):
    """最大回撤

    参考：https://blog.csdn.net/weixin_38997425/article/details/82915386

    :param n1b: 逐个结算周期的收益列表，单位：BP，换算关系是 10000BP = 100%
        如，n1b = [100.1, -90.5, 212.6]，表示第一个结算周期收益为100.1BP，也就是1.001%，以此类推。
    :return: 最大回撤起止位置和最大回撤
    """
    curve = np.cumsum(n1b)
    curve += 10000
    # 获取结束位置
    i = np.argmax((np.maximum.accumulate(curve) - curve) / np.maximum.accumulate(curve))
    if i == 0:
        return 0, 0, 0

    # 获取开始位置
    j = np.argmax(curve[:i])
    mdd = int((curve[j] - curve[i]) / curve[j] * 10000) / 10000
    return j, i, mdd


def turn_over_rate(df_holds: pd.DataFrame) -> [pd.DataFrame, float]:
    """计算持仓明细对应的组合换手率

    :param df_holds: 每个交易日的持仓明细，数据样例如下
                证券代码    成分日期    持仓权重
            0  000576.SZ  2020-01-02  0.0099
            1  000639.SZ  2020-01-02  0.0099
            2  000803.SZ  2020-01-02  0.0099
            3  000811.SZ  2020-01-02  0.0099
            4  000829.SZ  2020-01-02  0.0099
    :return: 组合换手率
    """
    dft = pd.pivot_table(df_holds, index='成分日期', columns='证券代码', values='持仓权重', aggfunc='sum')
    dft = dft.fillna(0)
    df_turns = dft.diff().abs().sum(axis=1).reset_index()
    df_turns.columns = ['date', 'change']

    # 由于是 diff 计算，第一个时刻的仓位变化被忽视了，修改一下
    sdt = df_holds['成分日期'].min()
    df_turns.loc[(df_turns['date'] == sdt), 'change'] = df_holds[df_holds['成分日期'] == sdt]['持仓权重'].sum()
    return df_turns, round(df_turns.change.sum() / 2, 4)


def holds_concepts_effect(holds: pd.DataFrame, concepts: dict, top_n=20, min_n=3, **kwargs):
    """股票持仓列表的板块效应

    原理概述：在选股时，如果股票的概念板块与组合中的其他股票的概念板块有重合，那么这个股票的表现会更好。

    :param holds: 组合股票池数据，样例：

            ===================  =========  ==========
            dt                   symbol         weight
            ===================  =========  ==========
            2023-05-09 00:00:00  601858.SH  0.00333333
            2023-05-09 00:00:00  300502.SZ  0.00333333
            2023-05-09 00:00:00  603258.SH  0.00333333
            2023-05-09 00:00:00  300499.SZ  0.00333333
            2023-05-09 00:00:00  300624.SZ  0.00333333
            ===================  =========  ==========

    :param concepts: 股票的概念板块，样例：
            {
                '002507.SZ': ['电子商务', '超级品牌', '国企改革'],
                '002508.SZ': ['家用电器', '杭州亚运会', '恒大概念']
            }
    :param top_n: 选取前 n 个密集概念
    :param min_n: 单股票至少要有 n 个概念在 top_n 中
    :return: 过滤后的选股结果，每个时间点的 top_n 概念
    """
    if kwargs.get('copy', True):
        holds = holds.copy()

    holds['概念板块'] = holds['symbol'].map(concepts).fillna('')
    holds['概念数量'] = holds['概念板块'].apply(len)
    holds = holds[holds['概念数量'] > 0]

    new_holds = []
    dt_key_concepts = {}
    for dt, dfg in tqdm(holds.groupby('dt'), desc='计算板块效应'):
        # 计算密集出现的概念
        key_concepts = [k for k, v in Counter([x for y in dfg['概念板块'] for x in y]).most_common(top_n)]
        dt_key_concepts[dt] = key_concepts

        # 计算在密集概念中出现次数超过min_n的股票
        dfg['强势概念'] = dfg['概念板块'].apply(lambda x: ','.join(set(x) & set(key_concepts)))
        sel = dfg[dfg['强势概念'].apply(lambda x: len(x.split(',')) >= min_n)]
        new_holds.append(sel)

    dfh = pd.concat(new_holds, ignore_index=True)
    dfk = pd.DataFrame([{"dt": k, '强势概念': v} for k, v in dt_key_concepts.items()])
    return dfh, dfk



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/7/7 22:42
describe: Event 相关的传感器
"""
import os
import shutil
import pandas as pd
from copy import deepcopy
from loguru import logger
from czsc.objects import Event
from typing import List, Dict, Callable, Any, Union
from czsc.traders.sig_parse import get_signals_freqs
from czsc.traders.base import generate_czsc_signals
from czsc.utils.io import save_json
from concurrent.futures import ProcessPoolExecutor, as_completed


class EventMatchSensor:
    def __init__(self, events: List[Union[Dict[str, Any], Event]], symbols: List[str], read_bars: Callable, **kwargs) -> None:
        """
        事件匹配传感器

        :param event: 事件配置
        :param symbols: 事件匹配的标的
        :param read_bars: 读取K线数据的函数，函数签名如下：
            read_bars(symbol, freq, sdt, edt, fq='前复权', **kwargs) -> List[RawBar]

        :param kwargs: 读取K线数据的函数的参数

            - bar_sdt: K线数据的开始时间，格式：2020-01-01
            - sdt: 事件匹配的开始时间，格式：2020-01-01
            - edt: 事件匹配的结束时间，格式：2020-01-01
            - max_workers: 读取K线数据的函数的最大进程数
            - signals_module: 信号解析模块，如：czsc.signals
            - results_path: 事件匹配结果的保存路径

        """
        self.symbols = symbols
        self.read_bars = read_bars
        self.events = [Event.load(event) if isinstance(event, dict) else event for event in events]
        self.events_map = {event.name: event for event in self.events}
        self.events_name = [event.name for event in self.events]
        self.results_path = kwargs.pop("results_path")
        self.refresh = kwargs.pop("refresh", False)
        if os.path.exists(self.results_path) and self.refresh:
            shutil.rmtree(self.results_path)
            logger.warning(f"文件夹 {self.results_path} 已存在，程序将覆盖该文件夹下的所有文件")
        os.makedirs(self.results_path, exist_ok=True)
        save_json({e.name: e.dump() for e in self.events}, os.path.join(self.results_path, "events.json"))

        logger.add(os.path.join(self.results_path, "event_match_sensor.log"), rotation="1 day", encoding="utf-8")
        logger.info(f"事件匹配传感器初始化，共有{len(self.events)}个事件，{len(self.symbols)}个标的")

        self.signals_module = kwargs.pop("signals_module", "czsc.signals")
        self.signals_config = self._get_signals_config()
        self.freqs = get_signals_freqs(self.signals_config)
        self.base_freq = self.freqs[0]
        logger.info(
            f"signals_moudle: {self.signals_module}, signals_config: {self.signals_config}, freqs: {self.freqs}"
        )

        self.bar_sdt = kwargs.pop("bar_sdt", "2017-01-01")
        self.sdt = kwargs.pop("sdt", "2018-01-01")
        self.edt = kwargs.pop("edt", "2022-01-01")
        logger.info(f"bar_sdt: {self.bar_sdt}, sdt: {self.sdt}, edt: {self.edt}")

        self.kwargs = kwargs
        logger.info(f"事件匹配传感器初始化完成，共有{len(self.events)}个事件，{len(self.symbols)}个标的")
        file_data = os.path.join(self.results_path, "data.feather")
        if os.path.exists(file_data):
            self.data = pd.read_feather(file_data)
            logger.info(f"读取事件匹配数据：{file_data}")
        else:
            self.data = self._multi_symbols(self.symbols, max_workers=self.kwargs.pop("max_workers", 1))
            self.data.to_feather(os.path.join(self.results_path, "data.feather"))

        _res = []
        for event_name in self.events_name:
            df = self.get_event_csc(event_name)
            df = df.set_index("dt")
            _res.append(df)
        df = pd.concat(_res, axis=1, ignore_index=False).reset_index()
        file_csc = os.path.join(self.results_path, "cross_section_counts.csv")
        df.to_csv(file_csc, index=False)
        logger.info(f"截面匹配次数计算完成，结果保存至：{file_csc}")

        # csc = cross section count，表示截面匹配次数
        self.csc = df

    def _get_signals_config(self):
        config = []
        for event in self.events:
            _c = event.get_signals_config(signals_module=self.signals_module)
            config.extend(_c)
        config = [dict(t) for t in {tuple(d.items()) for d in config}]
        return config

    def _single_symbol(self, symbol):
        """单个symbol的事件匹配"""
        try:
            bars = self.read_bars(symbol, freq=self.base_freq, sdt=self.bar_sdt, edt=self.edt, **self.kwargs)
            sigs = generate_czsc_signals(bars, deepcopy(self.signals_config), sdt=self.sdt, df=False)
            sigs = pd.DataFrame(sigs)
            events = deepcopy(self.events)
            new_cols = []
            for event in events:
                e_name = event.name
                sigs[[e_name, f'{e_name}_F']] = sigs.apply(event.is_match, axis=1, result_type="expand")  # type: ignore
                new_cols.extend([e_name, f'{e_name}_F'])
            sigs['n1b'] = (sigs['close'].shift(-1) / sigs['close'] - 1) * 10000
            sigs = sigs[['symbol', 'dt', 'open', 'close', 'high', 'low', 'vol', 'amount', 'n1b'] + new_cols]  # type: ignore
            return sigs
        except Exception as e:
            logger.error(f"{symbol} 事件匹配失败：{e}")
            return pd.DataFrame()

    def _multi_symbols(self, symbols: List[str], max_workers=1):
        """多个symbol的事件匹配"""
        logger.info(f"开始事件匹配，共有{len(symbols)}个标的，max_workers={max_workers}")
        dfs = []
        if max_workers == 1:
            for symbol in symbols:
                dfs.append(self._single_symbol(symbol))
        else:
            with ProcessPoolExecutor(max_workers) as executor:
                futures = [executor.submit(self._single_symbol, symbol) for symbol in symbols]
                for future in as_completed(futures):
                    dfs.append(future.result())
        df = pd.concat(dfs, ignore_index=True)
        for event_name in self.events_name:
            df[event_name] = df[event_name].astype(int)
        return df

    def get_event_csc(self, event_name: str):
        """获取事件的截面匹配次数

        csc = cross section count，表示截面匹配次数

        :param event_name: 事件名称
        :return: DataFrame
        """
        df = self.data.copy()
        df = df[df[event_name] == 1]
        df = df.groupby(["symbol", "dt"])[event_name].sum().reset_index()
        df = df.groupby("dt")[event_name].sum().reset_index()
        return df



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/1/29 15:01
describe: 相关系数计算、可视化

References:
1. https://zhuanlan.zhihu.com/p/362258222
2. https://blog.csdn.net/qq_45538220/article/details/107429201
"""
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn import metrics
from tqdm import tqdm
from typing import Union


plt.rcParams['font.sans-serif'] = ['SimHei']    # 用来正常显示中文标签
plt.rcParams['axes.unicode_minus'] = False      # 用来正常显示负号


def nmi_matrix(df: pd.DataFrame, heatmap=False) -> pd.DataFrame:
    """计算高维标准化互信息并以矩阵形式输出

    :param df: 数据
    :param heatmap: 是否绘制热力图
    :return:
    """
    cols = df.columns.to_list()

    m_dict = {}
    for i, col1 in tqdm(enumerate(cols), desc='nmi'):
        X = df[col1]
        for col2 in cols[i:]:
            Y = df[col2]
            nmi = metrics.normalized_mutual_info_score(X, Y)
            m_dict[f"{col1}_{col2}"] = nmi
            m_dict[f"{col2}_{col1}"] = nmi

    m = []
    for col1 in cols:
        A = []
        for col2 in cols:
            A.append(m_dict[f"{col1}_{col2}"])
        m.append(A)

    dfm = pd.DataFrame(m, index=cols, columns=cols)

    if heatmap:
        print('NMI(标准化互信息) = \n', dfm)
        plt.close()
        figure, ax = plt.subplots(figsize=(len(cols), len(cols)))
        sns.heatmap(dfm, square=True, annot=True, ax=ax)
        plt.show()
    return dfm


def single_linear(y: Union[np.array, list], x: Union[np.array, list] = None) -> dict:
    """单变量线性拟合

    :param y: 目标序列
    :param x: 单变量值
    :return res: 拟合结果，样例如下
        {'slope': 1.565, 'intercept': 67.9783, 'r2': 0.9967}

        slope       标识斜率
        intercept   截距
        r2          拟合优度
    """
    if not x:
        x = list(range(len(y)))

    x_squred_sum = sum([x1 * x1 for x1 in x])
    xy_product_sum = sum([x[i] * y[i] for i in range(len(x))])
    num = len(x)
    x_sum = sum(x)
    y_sum = sum(y)
    delta = float(num * x_squred_sum - x_sum * x_sum)
    if delta == 0:
        return {'slope': 0, 'intercept': 0, 'r2': 0}

    y_intercept = (1 / delta) * (x_squred_sum * y_sum - x_sum * xy_product_sum)
    slope = (1 / delta) * (num * xy_product_sum - x_sum * y_sum)

    y_mean = np.mean(y)
    ss_tot = sum([(y1 - y_mean) * (y1 - y_mean) for y1 in y]) + 0.00001
    ss_err = sum([(y[i] - slope * x[i] - y_intercept) * (y[i] - slope * x[i] - y_intercept) for i in range(len(x))])
    rsq = 1 - ss_err / ss_tot

    res = {'slope': round(slope, 4), 'intercept': round(y_intercept, 4), 'r2': round(rsq, 4)}
    return res


def cross_sectional_ic(df, x_col='open', y_col='n1b', method='spearman', **kwargs):
    """分析 df 中 x_col 和 y_col 列的截面相关性（IC）

    :param df：数据，DateFrame格式
    :param x_col：X列
    :param y_col：Y列，一般采用下期收益，也就是 n1b
    :param method：{'pearson', 'kendall', 'spearman'} or callable
            * pearson : standard correlation coefficient
            * kendall : Kendall Tau correlation coefficient
            * spearman : Spearman rank correlation
            * callable: callable with input two 1d ndarrays and returning a float
    :return：df，res: 前者是每日相关系数结果，后者是每日相关系数的统计结果
    """
    dt_col = kwargs.pop('dt_col', 'dt')
    tqdm.pandas(desc='cross_section_ic')
    s = df.groupby(dt_col).progress_apply(lambda row: row[x_col].corr(row[y_col], method=method))
    df = pd.DataFrame(s, columns=['ic']).reset_index(inplace=False)

    res = {
        "x_col": x_col,
        "y_col": y_col,
        "method": method,
        "IC均值": 0,
        "IC标准差": 0,
        "ICIR": 0,
        "IC胜率": 0,
        "IC绝对值>2%占比": 0,
    }
    if df.empty:
        return df, res

    df = df[~df['ic'].isnull()].copy()
    ic_avg = df['ic'].mean()
    ic_std = df['ic'].std()

    res['IC均值'] = round(ic_avg, 4)
    res['IC标准差'] = round(ic_std, 4)
    res['ICIR'] = round(ic_avg / ic_std, 4) if ic_std != 0 else 0
    res['IC胜率'] = round(len(df[df['ic'] > 0]) / len(df), 4)
    res['IC绝对值>2%占比'] = round(len(df[df['ic'].abs() > 0.02]) / len(df), 4)
    return df, res



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/5/11 20:36
describe: 信号观察页面
"""
import os
os.environ['czsc_max_bi_num'] = '20'
os.environ['czsc_research_cache'] = r"D:\CZSC投研数据"
import czsc
import numpy as np
import pandas as pd
import streamlit as st
from datetime import timedelta
from czsc.utils import sorted_freqs
from czsc.connectors.research import get_raw_bars, get_symbols

st.set_page_config(layout="wide")
signals_module = st.sidebar.text_input("信号模块名称：", value="czsc.signals")
parser = czsc.SignalsParser(signals_module=signals_module)

st.title("信号观察")

with st.sidebar:
    st.header("信号配置")
    with st.form("my_form"):
        conf = st.text_input("请输入信号：", value="60分钟_D0停顿分型_BE辅助V230106_看多_强_任意_0")
        symbol = st.selectbox("请选择股票：", get_symbols('ALL'), index=0)
        freqs = st.multiselect("请选择周期：", sorted_freqs, default=['30分钟', '日线', '周线'])
        freqs = czsc.freqs_sorted(freqs)
        sdt = st.date_input("开始日期：", value=pd.to_datetime('2022-01-01'))
        edt = st.date_input("结束日期：", value=pd.to_datetime('2023-01-01'))
        submit_button = st.form_submit_button(label='提交')


# 获取K线，计算信号
bars = get_raw_bars(symbol, freqs[0], pd.to_datetime(sdt) - timedelta(days=365*3), edt)
signals_config = czsc.get_signals_config([conf], signals_module=signals_module)
sigs = czsc.generate_czsc_signals(bars, signals_config, df=True, sdt=sdt)
sigs.drop(columns=['freq', 'cache'], inplace=True)
cols = [x for x in sigs.columns if len(x.split('_')) == 3]
assert len(cols) == 1
sigs['match'] = sigs.apply(czsc.Signal(conf).is_match, axis=1)
sigs['text'] = np.where(sigs['match'], sigs[cols[0]], "")

# 在图中绘制指定需要观察的信号
chart = czsc.KlineChart(n_rows=3, height=800)
chart.add_kline(sigs, freqs[0])
# 可以参考这里的代码，绘制其他自定义指标
chart.add_sma(sigs, row=1, ma_seq=(5, 10, 20), visible=True)
chart.add_vol(sigs, row=2)
chart.add_macd(sigs, row=3)
df1 = sigs[sigs['text'] != ""][['dt', 'text', 'close', 'low']].copy()
chart.add_scatter_indicator(x=df1['dt'], y=df1['low'], row=1, name='信号', mode='markers',
                            marker_size=20, marker_color='red', marker_symbol='triangle-up')
st.plotly_chart(chart.fig, use_container_width=True, config={
    "scrollZoom": True,
    "displayModeBar": True,
    "displaylogo": False,
    'modeBarButtonsToRemove': [
        'toggleSpikelines',
        'select2d',
        'zoomIn2d',
        'zoomOut2d',
        'lasso2d',
        'autoScale2d',
        'hoverClosestCartesian',
        'hoverCompareCartesian']})



import os
import czsc
import json
import glob
import hashlib
import pandas as pd
import streamlit as st
import plotly.express as px
from tqdm import tqdm
from loguru import logger
from czsc.utils.stats import net_value_stats
from datetime import timedelta
from multiprocessing import cpu_count
from czsc.connectors.research import get_symbols, get_raw_bars

st.set_page_config(layout="wide", page_title="CZSC策略回测", page_icon="🧊")

with st.sidebar:
    st.title("CZSC策略回测")
    form = st.form(key='my_form')
    files = form.file_uploader(label='上传策略文件', type='json', accept_multiple_files=True)
    symbol_gruop = form.selectbox(label="回测品类", options=['A股主要指数', 'A股场内基金', '中证500成分股', '期货主力'], index=3)
    start_date = form.date_input(label='开始日期', value=pd.to_datetime('2019-01-01'))
    end_date = form.date_input(label='结束日期', value=pd.to_datetime('2022-01-01'))
    max_workers = form.number_input(label='最大进程数', value=4, min_value=1, max_value=cpu_count() // 2)
    submit_button = form.form_submit_button(label='开始回测')

@st.cache_data()
def read_data(files_traders, pos_name):
    res = []
    for file in tqdm(files_traders):
        try:
            trader = czsc.dill_load(file)
            pos = trader.get_position(pos_name)

            hd = pd.DataFrame(pos.holds)
            hd['n1b'] = (hd['price'].shift(-1) / hd['price'] - 1) * 10000
            hd['edge_fee0'] = hd['pos'] * hd['n1b']
            hd['edge_fee2'] = hd['edge_fee0'] - hd['pos'].diff().abs() * 1
            hd['edge_fee4'] = hd['edge_fee0'] - hd['pos'].diff().abs() * 2

            hd['date'] = hd['dt'].dt.date
            daily = hd.groupby('date').agg({'edge_fee0': 'sum', 'edge_fee2': 'sum', 'edge_fee4': 'sum'}).reset_index()
            daily['symbol'] = trader.symbol
            res.append(daily)
        except Exception as e:
            logger.warning(f"{file} {pos_name} 读取失败: {e}")

    dfr = pd.concat(res, ignore_index=True)
    return dfr

if files and start_date and max_workers:
    strategies = {file.name: json.loads(file.getvalue().decode("utf-8")) for file in files}
    symbols = get_symbols(symbol_gruop)

    # 生成临时回测结果路径
    hash_code = hashlib.sha256(f"{str(strategies)}_{str(symbols)}".encode('utf-8')).hexdigest()[:8].upper()
    results_path = rf"D:\CTA投研\{symbol_gruop}_{start_date}_{end_date}_{hash_code}"

    if not os.path.exists(results_path):
        os.makedirs(results_path, exist_ok=True)
        os.makedirs(os.path.join(results_path, "upload_positions"), exist_ok=True)

        files_position = []
        for key, value in strategies.items():
            file_pos = os.path.join(results_path, "upload_positions", key)
            files_position.append(file_pos)
            czsc.save_json(value, file_pos)

        params = {
            "start_date": str(start_date),
            "end_date": str(end_date),
            "symbol_group": symbol_gruop,
            "symbols": symbols,
        }
        czsc.save_json(params, os.path.join(results_path, "params.json"))

        # 回测
        cta = czsc.CTAResearch(czsc.CzscJsonStrategy, get_raw_bars, results_path=results_path, 
                            files_position=files_position,
                            signals_module_name='czsc.signals')
        bar_sdt = pd.to_datetime(start_date) - timedelta(days=365)
        with st.spinner('正在执行策略回测，请耐心等候 ...'):
            cta.backtest(symbols, max_workers=int(max_workers), bar_sdt=bar_sdt, sdt=start_date, edt=end_date)

    # 生成回测报告
    file_traders = glob.glob(fr"{results_path}\backtest_*\traders\*.trader")

    st.subheader("一、品种等权收益曲线")
    all_pos_names = [x.name for x in czsc.dill_load(file_traders[0]).positions]
    pos_name = st.selectbox("选择持仓", all_pos_names, index=0)

    dfr = read_data(file_traders, pos_name)
    dfg = dfr.groupby('date').agg({'edge_fee0': 'mean', 'edge_fee2': 'mean', 'edge_fee4': 'mean'}).cumsum()
    dfg.rename({'edge_fee0': '等权费前收益', 'edge_fee2': '双边扣费2BP', 'edge_fee4': '双边扣费4BP'}, axis=1, inplace=True)

    fig = px.line(dfg, x=dfg.index, y=['等权费前收益', '双边扣费2BP', '双边扣费4BP'], labels=[], title="全部品种日收益等权")
    st.plotly_chart(fig, use_container_width=True, height=600)

    dfg['dt'] = dfg.index.to_list()
    stats = []
    for col in ['等权费前收益', '双边扣费2BP', '双边扣费4BP']:
        dfg_ = dfg[['dt', col]].copy().rename(columns={col: 'edge'}).reset_index(drop=True)
        dfg_['edge'] = dfg_['edge'].diff()
        stats_ = net_value_stats(dfg_, sub_cost=False)
        stats_['name'] = col
        stats.append(stats_)
    st.dataframe(pd.DataFrame(stats).set_index('name'), use_container_width=True)



from collections import OrderedDict
from czsc.analyze import CZSC
from czsc.objects import Direction, ZS
from czsc.signals.tas import update_macd_cache
from czsc.utils import create_single_signal, get_sub_elements


def tas_macd_bc_V230804(c: CZSC, **kwargs) -> OrderedDict:
    """MACD黄白线辅助背驰判断

    参数模板："{freq}_D{di}MACD背驰_BS辅助V230804"

    **信号逻辑：**

    以向上笔为例，当前笔在中枢中轴上方，且MACD黄白线不是最高，认为是背驰，做空；反之，做多。

    **信号列表：**

    - Signal('60分钟_D1MACD背驰_BS辅助V230804_空头_任意_任意_0')
    - Signal('60分钟_D1MACD背驰_BS辅助V230804_多头_任意_任意_0')

    :param c: CZSC对象
    :param kwargs: 无
    :return: 信号识别结果
    """
    di = int(kwargs.get('di', 1))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}MACD背驰_BS辅助V230804".split('_')
    v1 = '其他'
    cache_key = update_macd_cache(c)
    if len(c.bi_list) < 7 or len(c.bars_ubi) >= 7:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bis = get_sub_elements(c.bi_list, di=di, n=7)
    zs = ZS(bis=bis[-5:])
    if not zs.is_valid:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    dd = min([bi.low for bi in bis])
    gg = max([bi.high for bi in bis])
    b1, b2, b3, b4, b5 = bis[-5:]
    if b5.direction == Direction.Up and b5.high > (gg - (gg - dd) / 4):
        b5_dif = max([x.cache[cache_key]['dif'] for x in b5.fx_b.raw_bars])
        od_dif = max([x.cache[cache_key]['dif'] for x in b1.fx_b.raw_bars + b3.fx_b.raw_bars])
        if 0 < b5_dif < od_dif:
            v1 = '空头'
    
    if b5.direction == Direction.Down and b5.low < (dd + (gg - dd) / 4):
        b5_dif = min([x.cache[cache_key]['dif'] for x in b5.fx_b.raw_bars])
        od_dif = min([x.cache[cache_key]['dif'] for x in b1.fx_b.raw_bars + b3.fx_b.raw_bars])
        if 0 > b5_dif > od_dif:
            v1 = '多头'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def check():
    from czsc.connectors import research
    from czsc.traders.base import check_signals_acc

    symbols = research.get_symbols('A股主要指数')
    bars = research.get_raw_bars(symbols[0], '15分钟', '20181101', '20210101', fq='前复权')

    signals_config = [{'name': tas_macd_bc_V230804, 'freq': "60分钟"}]
    check_signals_acc(bars, signals_config=signals_config, height='780px', delta_days=5)  # type: ignore


if __name__ == '__main__':
    check()



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/11/15 20:45
describe: CZSC开源协作团队内部使用数据接口
"""
import os
import czsc
import pandas as pd
from tqdm import tqdm
from datetime import datetime
from czsc import RawBar, Freq

# 首次使用需要打开一个python终端按如下方式设置 token
# czsc.set_url_token(token='your token', url='http://zbczsc.com:9106')

cache_path = os.getenv("CZSC_CACHE_PATH", os.path.expanduser("~/.quant_data_cache"))
dc = czsc.DataClient(url='http://zbczsc.com:9106', cache_path=cache_path)


def format_kline(kline: pd.DataFrame, freq: Freq):
    """格式化K线数据

    :param kline: K线数据，格式如下：

        ==========  =========  ======  =======  ======  =====  ===========  ===========
        dt          code         open    close    high    low          vol       amount
        ==========  =========  ======  =======  ======  =====  ===========  ===========
        2022-01-04  600520.SH   20.54    21.12   21.17  20.33  2.1724e+06   1.94007e+07
        2022-01-05  600520.SH   21.17    20.73   21.29  20.52  1.8835e+06   1.67258e+07
        2022-01-06  600520.SH   20.56    21.17   21.57  18.69  3.4227e+06   3.11461e+07
        2022-01-07  600520.SH   21.5     20.61   21.5   20.61  2.51741e+06  2.24819e+07
        2022-01-10  600520.SH   20.4     21.69   21.69  20.4   4.80894e+06  4.39598e+07
        ==========  =========  ======  =======  ======  =====  ===========  ===========

    :return: 格式化后的K线数据
    """
    bars = []
    for i, row in kline.iterrows():
        bar = RawBar(symbol=row['code'], id=i, freq=freq, dt=row['dt'],
                     open=row['open'], close=row['close'], high=row['high'],
                     low=row['low'], vol=row['vol'], amount=row['amount'])
        bars.append(bar)
    return bars


def get_symbols(name, **kwargs):
    """获取指定分组下的所有标的代码

    :param name: 分组名称，可选值：'A股指数', 'ETF', '股票', '期货主力'
    :param kwargs:
    :return:
    """
    if name == "股票":
        data = dc.stock_basic(nobj=1, status=1)
        return data['code'].tolist()

    if name == "ETF":
        raise NotImplementedError

    if name == "A股指数":
        raise NotImplementedError

    if name == "期货主力":
        kline = dc.future_klines(trade_date="20231101")
        return kline['code'].unique().tolist()

    raise ValueError(f"{name} 分组无法识别，获取标的列表失败！")


def get_raw_bars(symbol, freq, sdt, edt, fq='前复权', **kwargs):
    """获取 CZSC 库定义的标准 RawBar 对象列表

    :param symbol: 标的代码
    :param freq: 周期，支持 Freq 对象，或者字符串，如
            '1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线', '周线', '月线', '季线', '年线'
    :param sdt: 开始时间
    :param edt: 结束时间
    :param fq: 除权类型，可选值：'前复权', '后复权', '不复权'
    :param kwargs:
    :return:
    """
    freq = czsc.Freq(freq)

    if symbol.endswith(".SH") or symbol.endswith(".SZ"):
        fq_map = {"前复权": "qfq", "后复权": "hfq", "不复权": None}
        adj = fq_map.get(fq, None)
        if freq.value.endswith('分钟'):
            df = dc.pro_bar(code=symbol, sdt=sdt, edt=edt, freq='min', adj=adj)
            df = df[~df['dt'].str.endswith("09:30:00")].reset_index(drop=True)
        else:
            df = dc.pro_bar(code=symbol, sdt=sdt, edt=edt, freq='day', adj=adj)
        df.rename(columns={'code': 'symbol'}, inplace=True)
        df['dt'] = pd.to_datetime(df['dt'])
        return czsc.resample_bars(df, target_freq=freq)

    if symbol.endswith("9001"):
        if freq.value.endswith('分钟'):
            df = dc.future_klines(code=symbol, sdt=sdt, edt=edt, freq='1m')
        else:
            df = dc.future_klines(code=symbol, sdt=sdt, edt=edt, freq='1d')
        df.rename(columns={'code': 'symbol'}, inplace=True)
        df['amount'] = df['vol'] * df['close']
        df = df[['symbol', 'dt', 'open', 'close', 'high', 'low', 'vol', 'amount']].copy().reset_index(drop=True)
        df['dt'] = pd.to_datetime(df['dt'])
        return czsc.resample_bars(df, target_freq=freq)

    raise ValueError(f"symbol {symbol} 无法识别，获取数据失败！")


def stocks_daily_klines(years=None, **kwargs):
    """获取全市场A股的日线数据"""
    adj = kwargs.get('adj', 'hfq')
    if years is None:
        years = ['2017', '2018', '2019', '2020', '2021', '2022', '2023']

    res = []
    for year in years:
        ttl = 3600 * 6 if year == str(datetime.now().year) else -1
        kline = dc.pro_bar(trade_year=year, adj=adj, v=2, ttl=ttl)
        res.append(kline)

    dfk = pd.concat(res, ignore_index=True)
    dfk['dt'] = pd.to_datetime(dfk['dt'])
    dfk = dfk.sort_values(['code', 'dt'], ascending=True).reset_index(drop=True)
    if kwargs.get('exclude_bj', True):
        dfk = dfk[~dfk['code'].str.endswith(".BJ")].reset_index(drop=True)

    nxb = kwargs.get('nxb', [1, 2, 5])
    if nxb:
        rows = []
        for _, dfg in tqdm(dfk.groupby('code'), desc="计算NXB收益率", ncols=80, colour='green'):
            czsc.update_nbars(dfg, numbers=nxb, move=1, price_col='open')
            rows.append(dfg)
        dfk = pd.concat(rows, ignore_index=True)

    dfk = dfk.rename(columns={'code': 'symbol'})
    return dfk



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/21 16:09
describe: 测试交易价格计算
"""
import czsc
import numpy as np
from test.test_analyze import read_1min


def test_trade_price():
    bars = read_1min()
    df = czsc.cal_trade_price(bars)

    assert df['TWAP20'].iloc[0] == round(df['close'].iloc[1:21].mean(), 3)
    close = df['close'].iloc[1:21]
    vol = df['vol'].iloc[1:21]
    assert df['VWAP20'].iloc[0] == round(np.average(close, weights=vol), 3)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/7/15 13:42
describe: 特征分析相关的传感器
"""
import os
import pandas as pd
from tqdm import tqdm
from loguru import logger
from czsc.utils.corr import cross_sectional_ic
from czsc.utils.stats import daily_performance
from czsc.utils.trade import update_nbars
from concurrent.futures import ProcessPoolExecutor, as_completed


class FeatureAnalyzeBase:
    """【基类】特征计算与分析，适用于时序量价类因子"""

    def __init__(self, symbols, read_bars, **kwargs) -> None:
        """初始化函数

        :param symbols: list，需要获取特征的品种列表
        :param read_bars: function，获取品种K线数据的函数
        :param kwargs: dict，其他参数

            - freq: str，K线周期，可选值：日线、60分钟、30分钟、15分钟、5分钟、1分钟
            - sdt: str，开始日期
            - edt: str，结束日期
            - fq: str，复权方式，可选值：前复权、后复权，默认为后复权
            - max_workers: int，多进程获取特征的最大进程数
        """
        self.symbols = symbols
        self.read_bars = read_bars
        self.kwargs = kwargs
        self.dfs = self.get_features()
        self.report()

    @property
    def new_features(self):
        """list，新增的特征列表"""
        raise NotImplementedError

    def add_features(self, df):
        """向df中添加特征

        df 包含以下列：

        - dt: 日期
        - open: 开盘价
        - close: 收盘价
        - high: 最高价
        - low: 最低价
        - vol: 成交量
        - amount: 成交额
        """
        raise NotImplementedError

    def _one_symbol_features(self, symbol):
        """获取单个品种的特征"""
        freq = self.kwargs.get('freq', '日线')
        sdt = self.kwargs.get('sdt', '2010-01-01')
        edt = self.kwargs.get('edt', '2023-01-01')
        fq = self.kwargs.get('fq', "后复权")
        nbars_seq = self.kwargs.get('nbars_seq', (1, 2, 5))
        try:
            bars = self.read_bars(symbol=symbol, freq=freq, sdt=sdt, edt=edt, fq=fq)
            _df = pd.DataFrame(bars)
            update_nbars(_df, numbers=nbars_seq, price_col='open', move=1)
            _df = self.add_features(_df)
            return _df
        except Exception as e:
            logger.error(f"{symbol} error: {e}")
            return pd.DataFrame()

    def get_features(self) -> pd.DataFrame:
        """
        此方法检索self.symbols中每个符号的K线数据，
        使用add_features方法向数据添加特征

        :return: 包含所有K线和特征的DataFrame
        """
        max_workers = self.kwargs.get('max_workers', 1)
        if max_workers == 1:
            res = []
            for symbol in tqdm(self.symbols, desc="获取特征"):
                df = self._one_symbol_features(symbol)
                res.append(df)
        else:
            res = []
            with ProcessPoolExecutor(max_workers) as executor:
                tasks = [executor.submit(self._one_symbol_features, symbol) for symbol in self.symbols]
                for future in tqdm(as_completed(tasks), desc="获取特征", total=len(tasks)):
                    df = future.result()
                    res.append(df)

        return pd.concat(res, ignore_index=True)

    def layering(self, feature, min_q, max_q):
        """分层回测"""
        df = self.dfs.copy()
        # q 是截面排序，按照特征值从大到小排序，q的取值范围是[0, 1]
        df['q'] = df.groupby('dt')[feature].transform(lambda x: x.rank(pct=True))
        df1 = df[(df['q'] >= min_q) & (df['q'] <= max_q)]
        dfm = df1.groupby('dt').agg({'n1b': 'mean'}).fillna(0)
        logger.info(f"分层累计收益 {feature} {min_q}-{max_q}：{dfm['n1b'].sum():.4f}; {daily_performance((dfm['n1b'] / 10000).to_list())}")
        return dfm

    def report(self):
        """打印特征分析报告"""
        results_path = self.kwargs.get('results_path', None)
        if results_path:
            os.makedirs(results_path, exist_ok=True)
            dfs = self.dfs.copy()
            dfs['freq'] = dfs['freq'].apply(lambda x: x.value)
            dfs.drop(['cache'], axis=1, inplace=True)
            logger.add(os.path.join(results_path, "feature.log"), rotation="1 week", encoding='utf-8')
            dfs.to_feather(os.path.join(results_path, "features.feather"))

        logger.info(f"新增特征：{self.new_features}")
        corr_method = self.kwargs.get('corr_method', 'pearson')
        logger.info(f"相关系数计算方法：{corr_method}")

        for feature in self.new_features:
            logger.info(f"特征 {feature} 的取值范围：{self.dfs[feature].describe().round(4).to_dict()}")
            df1, res1 = cross_sectional_ic(self.dfs, x_col=feature, y_col='n1b', method=corr_method, dt_col='dt')
            logger.info(f"特征 {feature} 与未来1日收益的相关系数：{res1}")
            _ = self.layering(feature, 0.95, 1)
            _ = self.layering(feature, 0.9, 1)
            _ = self.layering(feature, 0, 0.1)
            _ = self.layering(feature, 0, 0.05)

            df1['年月'] = df1['dt'].apply(lambda x: x.strftime("%Y年%m月"))
            dfm = df1.groupby('年月').agg({'ic': 'mean'})
            logger.info(f"特征 {feature} 与未来1日收益的相关系数月度描述：{dfm.describe().round(4).to_dict()}")
            logger.info(f"特征 {feature} 与未来1日收益的相关系数月度胜率：{len(dfm[dfm['ic'] > 0])/len(dfm):.4f}")


class FixedNumberSelector:
    """选择固定数量（等权）的交易品种

    可优化项：
    1. 传入 res_path, 将分析过程和分析结果保存下来
    2. 支持传入大盘择时信号，例如：大盘择时信号为空头时，多头只平不开
    """

    def __init__(self, dfs, k, d, **kwargs):
        """

        :param dfs: pd.DataFrame，所有交易品种的特征打分数据，必须包含以下列：dt, open, close, high, low, vol, amount, score；数据样例：

            ===================  =========  =======  =======  =======  =======  ========  =========  ========  =============
            dt                   symbol        open    close     high      low       vol     amount       n1b          score
            ===================  =========  =======  =======  =======  =======  ========  =========  ========  =============
            2017-01-03 00:00:00  000001.SZ  954.345  959.583  961.678  952.25   45984049  420595176   21.8583  nan
            2017-01-04 00:00:00  000001.SZ  958.536  959.583  961.678  957.488  44932953  411503444    0         0
            2017-01-05 00:00:00  000001.SZ  960.631  960.631  961.678  958.536  34437291  315769693  -43.6213    3.17018e-11
            2017-01-06 00:00:00  000001.SZ  960.631  956.441  960.631  954.345  35815420  327176433   21.9062   -1.21795e-10
            2017-01-09 00:00:00  000001.SZ  956.441  958.536  960.631  954.345  36108157  329994604  -10.9292    6.06684e-11
            2017-01-10 00:00:00  000001.SZ  958.536  958.536  959.583  957.488  24105395  220575131  -10.9411    0
            2017-01-11 00:00:00  000001.SZ  957.488  957.488  960.631  956.441  30343089  277553207   10.9531   -3.60186e-11
            2017-01-12 00:00:00  000001.SZ  956.441  958.536  960.631  956.441  42800677  391869402   10.9411    2.5563e-11
            2017-01-13 00:00:00  000001.SZ  957.488  959.583  962.726  955.393  43430137  397601906  -32.7865    2.51649e-11
            2017-01-16 00:00:00  000001.SZ  958.536  957.488  959.583  950.155  68316586  623025820   21.9292   -3.19607e-11
            ===================  =========  =======  =======  =======  =======  ========  =========  ========  =============

        :param k: int，每期固定选择的数量
        :param d: int，每期允许变动的数量
        """
        logger.info(f"选择固定数量的交易品种，k={k}，d={d}, dfs.shape={dfs.shape}, kwargs={kwargs}")
        self.dfs = dfs   # 所有交易品种的特征打分数据，必须包含以下列：dt, open, close, high, low, vol, amount, score
        self.k = k       # 每期固定选择的数量
        self.d = d       # 每期允许变动的数量
        self.kwargs = kwargs
        self.is_stocks = kwargs.get('is_stocks', False)  # 是否是A股，如果是A股，需要考虑涨跌停的情况
        self.__preprocess()

        self.operate_fee = kwargs.get('operate_fee', 15)  # 单边手续费+交易滑点，单位：BP
        self.holds = {}  # 每期持有的品种
        self.operates = {}  # 每期操作的品种
        for dt in self.dts:
            self.__deal_one_time(dt)

    def __preprocess(self):
        assert 'dt' in self.dfs.columns, "必须包含dt列"
        assert 'n1b' in self.dfs.columns, "必须包含n1b列"
        assert 'symbol' in self.dfs.columns, "必须包含symbol列"
        assert 'score' in self.dfs.columns, "必须包含score列, 这是选择交易品种的依据"

        self.dfs['dt'] = pd.to_datetime(self.dfs['dt']).dt.strftime("%Y-%m-%d %H:%M:%S")
        dts = sorted(self.dfs['dt'].unique())
        last_dt_map = {dt: dts[i-1] for i, dt in enumerate(dts)}
        self.dts, self.last_dt_map = dts, last_dt_map
        self.score_map = {dt: dfg[['symbol', 'dt', 'open', 'close', 'high', 'low', 'score', 'n1b']].copy() for dt, dfg in self.dfs.groupby('dt')}

    def __deal_one_time(self, dt):
        """单次调整记录"""
        k, d, is_stocks = self.k, self.d, self.is_stocks

        score = self.score_map[dt]
        if is_stocks:
            zt_symbols = [x['symbol'] for _, x in score.iterrows() if x['close'] == x['high'] >= x['open']]
            dt_symbols = [x['symbol'] for _, x in score.iterrows() if x['close'] == x['low'] <= x['open']]
            score_a = score[~score.symbol.isin(zt_symbols + dt_symbols)].copy()
            logger.info(f"A股今日{dt}涨停{len(zt_symbols)}个品种，跌停{len(dt_symbols)}个品种，已跳过")
        else:
            score_a = score.copy()

        if not self.holds:
            logger.info(f"当前持仓为空，选择前{k}个品种")
            assert not self.operates, "当holds是空的时候，操作记录必须为空"

            _df = score_a.sort_values(by='score', ascending=False).head(k)
            _df['edge'] = _df['n1b'] - self.operate_fee
            self.holds[dt] = _df

            _df_operates = [{'symbol': row['symbol'], 'dt': dt, 'action': 'buy', 'price': row['close']} for _, row in _df.iterrows()]
            self.operates[dt] = pd.DataFrame(_df_operates)
            return

        # 有持仓的情况
        score = self.score_map[dt]
        last_dt = self.last_dt_map[dt]
        last_holds = self.holds[last_dt].copy()
        last_symbols = last_holds['symbol'].tolist()
        skip_symbols = [x for x in last_symbols if x not in score['symbol'].tolist()]
        if skip_symbols:
            logger.warning(f"【数据缺陷提示】上一期持仓中，有{len(skip_symbols)}个品种，本期{dt}不在交易品种中，已跳过: {skip_symbols}")

        topk_symbols = score_a.sort_values(by='score', ascending=False).head(k)['symbol'].tolist()
        sell_symbols = score_a[score_a.symbol.isin(last_symbols)].sort_values(by='score', ascending=False).tail(d)['symbol'].tolist()
        sell_symbols = [x for x in sell_symbols if x not in topk_symbols] + skip_symbols
        keep_symbols = [x for x in last_symbols if x not in sell_symbols]
        if len(keep_symbols) != k - len(sell_symbols):
            logger.warning(f"保持品种数量不对，当前只有{len(keep_symbols)}个品种")

        buy_symbols = score_a[~score_a.symbol.isin(keep_symbols)].sort_values(by='score', ascending=False).head(len(sell_symbols))['symbol'].tolist()
        assert len(buy_symbols) == len(sell_symbols), "买入品种数量必须等于卖出品种数量"
        assert len(keep_symbols + buy_symbols) == k, "保持品种数量+买入品种数量必须等于k"
        _df = score[score.symbol.isin(keep_symbols + buy_symbols)].sort_values(by='score', ascending=False)

        if len(_df) != k:
            logger.warning(f"选择的品种数量不等于{k}，当前只有{len(_df)}个品种")

        _df['edge'] = _df.apply(lambda row: row['n1b'] - self.operate_fee if row['symbol'] in buy_symbols else row['n1b'], axis=1)
        self.holds[dt] = _df

        # 平仓扣费，在上一期的持仓中，卖出的品种，需要扣除手续费
        last_holds['edge'] = last_holds.apply(lambda row: row['edge'] - self.operate_fee if row['symbol'] in sell_symbols else row['edge'], axis=1)
        self.holds[last_dt] = last_holds

        _sell_operates = [{'symbol': row['symbol'], 'dt': dt, 'action': 'sell', 'price': row['close']}
                          for _, row in score[score.symbol.isin(sell_symbols)].iterrows()]
        _buy_operates = [{'symbol': row['symbol'], 'dt': dt, 'action': 'buy', 'price': row['close']}
                         for _, row in score[score.symbol.isin(buy_symbols)].iterrows()]
        _df_operates = pd.DataFrame(_sell_operates + _buy_operates)
        self.operates[dt] = _df_operates



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/5 20:45
describe: CZSC投研数据共享接口

投研数据共享说明（含下载地址）：https://s0cqcxuy3p.feishu.cn/wiki/wikcnzuPawXtBB7Cj7mqlYZxpDh
"""
import os
import czsc
import glob
import pandas as pd

# 投研共享数据的本地缓存路径，需要根据实际情况修改
cache_path = os.environ.get('czsc_research_cache', r"D:\CZSC投研数据")
if not os.path.exists(cache_path):
    raise ValueError(f"请设置环境变量 czsc_research_cache 为投研共享数据的本地缓存路径，当前路径不存在：{cache_path}。\n\n"
                     f"投研数据共享说明（含下载地址）：https://s0cqcxuy3p.feishu.cn/wiki/wikcnzuPawXtBB7Cj7mqlYZxpDh")


def get_symbols(name, **kwargs):
    """获取指定分组下的所有标的代码

    :param name: 分组名称，可选值：'A股主要指数', 'A股场内基金', '中证500成分股', '期货主力'
    :param kwargs:
    :return:
    """
    if name.upper() == 'ALL':
        files = glob.glob(os.path.join(cache_path, "*", "*.parquet"))
    else:
        files = glob.glob(os.path.join(cache_path, name, "*.parquet"))
    return [os.path.basename(x).replace('.parquet', '') for x in files]


def get_raw_bars(symbol, freq, sdt, edt, fq='前复权', **kwargs):
    """获取 CZSC 库定义的标准 RawBar 对象列表

    :param symbol: 标的代码
    :param freq: 周期，支持 Freq 对象，或者字符串，如
            '1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线', '周线', '月线', '季线', '年线'
    :param sdt: 开始时间
    :param edt: 结束时间
    :param fq: 除权类型，投研共享数据默认都是后复权，不需要再处理
    :param kwargs:
    :return:
    """
    kwargs['fq'] = fq
    file = glob.glob(os.path.join(cache_path, "*", f"{symbol}.parquet"))[0]
    freq = czsc.Freq(freq)
    kline = pd.read_parquet(file)
    if 'dt' not in kline.columns:
        kline['dt'] = pd.to_datetime(kline['datetime'])
    kline = kline[(kline['dt'] >= pd.to_datetime(sdt)) & (kline['dt'] <= pd.to_datetime(edt))]
    if kline.empty:
        return []
    _bars = czsc.resample_bars(kline, freq, raw_bars=True, base_freq='1分钟')
    return _bars



import czsc
import numpy as np
import pandas as pd
import streamlit as st
import plotly.express as px
from sklearn.linear_model import LinearRegression


def show_daily_return(df, **kwargs):
    """用 streamlit 展示日收益"""
    assert df.index.dtype == 'datetime64[ns]', "index必须是datetime64[ns]类型, 请先使用 pd.to_datetime 进行转换"
    df = df.copy().fillna(0)

    def _stats(df_, type_='持有日'):
        df_ = df_.copy()
        stats = []
        for col in df_.columns:
            if type_ == '持有日':
                col_stats = czsc.daily_performance([x for x in df_[col] if x != 0])
            else:
                assert type_ == '交易日', "type_ 参数必须是 持有日 或 交易日"
                col_stats = czsc.daily_performance(df_[col])
            col_stats['日收益名称'] = col
            stats.append(col_stats)
        stats = pd.DataFrame(stats).set_index('日收益名称')
        fmt_cols = ['年化', '夏普', '最大回撤', '卡玛', '年化波动率', '非零覆盖', '日胜率', '盈亏平衡点']
        stats = stats.style.background_gradient(cmap='RdYlGn_r', axis=None, subset=fmt_cols).format('{:.4f}')
        return stats

    with st.container():
        title = kwargs.get("title", "")
        if title:
            st.subheader(title)
            st.divider()

        col1, col2 = st.columns([1, 1])
        col1.write("交易日绩效指标")
        col1.dataframe(_stats(df, type_='交易日'), use_container_width=True)
        col2.write("持有日绩效指标")
        col2.dataframe(_stats(df, type_='持有日'), use_container_width=True)

        df = df.cumsum()
        fig = px.line(df, y=df.columns.to_list(), title="日收益累计曲线")
        fig.update_xaxes(title='')

        # 添加每年的开始第一个日期的竖线
        for year in range(df.index.year.min(), df.index.year.max() + 1):
            first_date = df[df.index.year == year].index.min()
            fig.add_vline(x=first_date, line_dash='dash', line_color='red')

        for col in kwargs.get("legend_only_cols", []):
            fig.update_traces(visible="legendonly", selector=dict(name=col))

        st.plotly_chart(fig, use_container_width=True)


def show_correlation(df, cols=None, method='pearson', **kwargs):
    """用 streamlit 展示相关性

    :param df: pd.DataFrame，数据源
    :param cols: list，分析相关性的字段
    :param method: str，计算相关性的方法，可选 pearson 和 spearman
    """
    cols = cols or df.columns.to_list()
    dfr = df[cols].corr(method=method)
    dfr['average'] = (dfr.sum(axis=1) - 1) / (len(cols) - 1)
    dfr = dfr.style.background_gradient(cmap='RdYlGn_r', axis=None).format('{:.4f}', na_rep='MISS')
    st.dataframe(dfr, use_container_width=kwargs.get("use_container_width", True))


def show_sectional_ic(df, x_col, y_col, method='pearson', **kwargs):
    """使用 streamlit 展示截面IC

    :param df: pd.DataFrame，数据源
    :param x_col: str，因子列名
    :param y_col: str，收益列名
    :param method: str，计算IC的方法，可选 pearson 和 spearman
    """
    dfc, res = czsc.cross_sectional_ic(df, x_col=x_col, y_col=y_col, dt_col='dt', method=method)

    col1, col2, col3, col4 = st.columns([1, 1, 1, 5])
    col1.metric("IC均值", res['IC均值'])
    col1.metric("IC标准差", res['IC标准差'])
    col2.metric("ICIR", res['ICIR'])
    col2.metric("IC胜率", res['IC胜率'])
    col3.metric("IC绝对值>2%占比", res['IC绝对值>2%占比'])
    col3.metric("品种数量", df['symbol'].nunique())

    dfc[['year', 'month']] = dfc.dt.apply(lambda x: pd.Series([x.year, x.month]))
    dfm = dfc.groupby(['year', 'month']).agg({'ic': 'mean'}).reset_index()
    dfm = pd.pivot_table(dfm, index='year', columns='month', values='ic')

    col4.write("月度IC分析结果：")
    col4.dataframe(dfm.style.background_gradient(cmap='RdYlGn_r', axis=None).format('{:.4f}', na_rep='MISS'),
                   use_container_width=True)

    fig = px.histogram(df, x=x_col, marginal="box", title="因子数据分布图")
    st.plotly_chart(fig, use_container_width=True)


def show_factor_returns(df, x_col, y_col):
    """使用 streamlit 展示因子收益率

    :param df: pd.DataFrame，数据源
    :param x_col: str，因子列名
    :param y_col: str，收益列名
    """
    assert 'dt' in df.columns, "时间列必须为 dt"

    res = []
    for dt, dfg in df.groupby("dt"):
        dfg = dfg.copy().dropna(subset=[x_col, y_col])
        X = dfg[x_col].values.reshape(-1, 1)
        y = dfg[y_col].values.reshape(-1, 1)
        model = LinearRegression(fit_intercept=False).fit(X, y)
        res.append([dt, model.coef_[0][0]])

    res = pd.DataFrame(res, columns=["dt", "因子收益率"])
    res['dt'] = pd.to_datetime(res['dt'])

    col1, col2 = st.columns(2)
    fig = px.bar(res, x='dt', y="因子收益率", title="因子逐K收益率")
    col1.plotly_chart(fig, use_container_width=True)

    res["因子累计收益率"] = res["因子收益率"].cumsum()
    fig = px.line(res, x='dt', y="因子累计收益率", title="因子累计收益率")
    col2.plotly_chart(fig, use_container_width=True)


def show_factor_layering(df, x_col, y_col='n1b', **kwargs):
    """使用 streamlit 绘制因子分层收益率图

    :param df: 因子数据
    :param x_col: 因子列名
    :param y_col: 收益列名
    :param kwargs:

        - n: 分层数量，默认为10
        - long: 多头组合，例如 "第10层"
        - short: 空头组合，例如 "第01层"

    """
    n = kwargs.get("n", 10)
    if df[y_col].max() > 100:       # 收益率单位为BP, 转换为万分之一
        df[y_col] = df[y_col] / 10000

    df = czsc.feture_cross_layering(df, x_col, n=n)

    mr = df.groupby(["dt", f'{x_col}分层'])[y_col].mean().reset_index()
    mrr = mr.pivot(index='dt', columns=f'{x_col}分层', values=y_col).fillna(0)

    tabs = st.tabs(["分层收益率", "多空组合"])
    with tabs[0]:
        czsc.show_daily_return(mrr)

    with tabs[1]:
        layering_cols = mrr.columns.to_list()
        with st.form(key="factor_form"):
            col1, col2 = st.columns(2)
            long = col1.multiselect("多头组合", layering_cols, default=[], key="factor_long")
            short = col2.multiselect("空头组合", layering_cols, default=[], key="factor_short")
            submit = st.form_submit_button("多空组合快速测试")

        if not submit:
            st.warning("请设置多空组合")
            st.stop()

        dfr = mrr.copy()
        dfr['多头'] = dfr[long].mean(axis=1)
        dfr['空头'] = -dfr[short].mean(axis=1)
        dfr['多空'] = (dfr['多头'] + dfr['空头']) / 2
        czsc.show_daily_return(dfr[['多头', '空头', '多空']])


def show_symbol_factor_layering(df, x_col, y_col='n1b', **kwargs):
    """使用 streamlit 绘制单个标的上的因子分层收益率图

    :param df: 因子数据，必须包含 dt, x_col, y_col 列，其中 dt 为日期，x_col 为因子值，y_col 为收益率，数据样例：

        ===================  ============  ============
        dt                      intercept     n1b
        ===================  ============  ============
        2017-01-03 00:00:00   0             0.00716081
        2017-01-04 00:00:00  -0.00154541    0.000250816
        2017-01-05 00:00:00   0.000628884  -0.0062695
        2017-01-06 00:00:00  -0.00681021    0.00334212
        2017-01-09 00:00:00   0.00301077   -0.00182963
        ===================  ============  ============

    :param x_col: 因子列名
    :param y_col: 收益列名
    :param kwargs:

        - n: 分层数量，默认为10

    """
    df = df.copy()
    n = kwargs.get("n", 10)
    if df[y_col].max() > 100:       # 如果收益率单位为BP, 转换为万分之一
        df[y_col] = df[y_col] / 10000

    if f'{x_col}分层' not in df.columns:
        # 如果因子分层列不存在，先计算因子分层
        if df[x_col].nunique() > n:
            czsc.normalize_ts_feature(df, x_col, n=n)
        else:
            # 如果因子值的取值数量小于分层数量，直接使用因子独立值排序作为分层
            x_rank = sorted(df[x_col].unique())
            x_rank = {x_rank[i]: f'第{str(i+1).zfill(2)}层' for i in range(len(x_rank))}
            st.success(f"因子值分层对应关系：{x_rank}")
            df[f'{x_col}分层'] = df[x_col].apply(lambda x: x_rank[x])

    for i in range(n):
        df[f'第{str(i+1).zfill(2)}层'] = np.where(df[f'{x_col}分层'] == f'第{str(i+1).zfill(2)}层', df[y_col], 0)

    layering_cols = [f'第{str(i).zfill(2)}层' for i in range(1, n + 1)]
    mrr = df[['dt'] + layering_cols].copy()
    mrr.set_index('dt', inplace=True)

    tabs = st.tabs(["分层收益率", "多空组合"])

    with tabs[0]:
        show_daily_return(mrr)

    with tabs[1]:
        col1, col2 = st.columns(2)
        long = col1.multiselect("多头组合", layering_cols, default=["第02层"], key="symbol_factor_long")
        short = col2.multiselect("空头组合", layering_cols, default=["第01层"], key="symbol_factor_short")
        dfr = mrr.copy()
        dfr['多头'] = dfr[long].sum(axis=1)
        dfr['空头'] = -dfr[short].sum(axis=1)
        dfr['多空'] = dfr['多头'] + dfr['空头']
        show_daily_return(dfr[['多头', '空头', '多空']])


def show_weight_backtest(dfw, **kwargs):
    """展示权重回测结果

    :param dfw: 回测数据，任何字段都不允许有空值；数据样例：

        ===================  ========  ========  =======
        dt                   symbol      weight    price
        ===================  ========  ========  =======
        2019-01-02 09:01:00  DLi9001       0.5   961.695
        2019-01-02 09:02:00  DLi9001       0.25  960.72
        2019-01-02 09:03:00  DLi9001       0.25  962.669
        2019-01-02 09:04:00  DLi9001       0.25  960.72
        2019-01-02 09:05:00  DLi9001       0.25  961.695
        ===================  ========  ========  =======

    :param kwargs:

        - fee: 单边手续费，单位为BP，默认为2BP
    """
    fee = kwargs.get("fee", 2)
    digits = kwargs.get("digits", 2)
    if (dfw.isnull().sum().sum() > 0) or (dfw.isna().sum().sum() > 0):
        st.warning("show_weight_backtest :: 持仓权重数据中存在空值，请检查数据后再试；空值数据如下：")
        st.dataframe(dfw[dfw.isnull().sum(axis=1) > 0], use_container_width=True)
        st.stop()

    from czsc.traders.weight_backtest import WeightBacktest

    wb = WeightBacktest(dfw, fee_rate=fee / 10000, digits=digits)
    stat = wb.results['绩效评价']

    st.divider()
    c1, c2, c3, c4, c5, c6, c7, c8 = st.columns([1, 1, 1, 1, 1, 1, 1, 1])
    c1.metric("盈亏平衡点", f"{stat['盈亏平衡点']:.2%}")
    c2.metric("单笔收益", f"{stat['单笔收益']} BP")
    c3.metric("交易胜率", f"{stat['交易胜率']:.2%}")
    c4.metric("持仓K线数", f"{stat['持仓K线数']}")
    c5.metric("最大回撤", f"{stat['最大回撤']:.2%}")
    c6.metric("年化收益率", f"{stat['年化']:.2%}")
    c7.metric("夏普比率", f"{stat['夏普']:.2f}")
    c8.metric("卡玛比率", f"{stat['卡玛']:.2f}")
    st.divider()

    dret = wb.results['品种等权日收益']
    dret.index = pd.to_datetime(dret.index)
    show_daily_return(dret, legend_only_cols=dfw['symbol'].unique().tolist())
    return wb
